# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
functions for digesting paths
"""
from __future__ import annotations

import re
import copy
import abc

from math import atan2, cos, pi, sin, sqrt, acos, tan
from typing import (
    Any,
    Type,
    Dict,
    Optional,
    Union,
    Tuple,
    List,
    Generator,
    TypeVar,
)
from .transforms import (
    Transform,
    BoundingBox,
    Vector2d,
    cubic_extrema,
    quadratic_extrema,
)
from .utils import classproperty, strargs


Pathlike = TypeVar("Pathlike", bound="PathCommand")
AbsolutePathlike = TypeVar("AbsolutePathlike", bound="AbsolutePathCommand")

# All the names that get added to the inkex API itself.
__all__ = (
    "Path",
    "CubicSuperPath",
    "PathCommand",
    "AbsolutePathCommand",
    "RelativePathCommand",
    # Path commands:
    "Line",
    "line",
    "Move",
    "move",
    "ZoneClose",
    "zoneClose",
    "Horz",
    "horz",
    "Vert",
    "vert",
    "Curve",
    "curve",
    "Smooth",
    "smooth",
    "Quadratic",
    "quadratic",
    "TepidQuadratic",
    "tepidQuadratic",
    "Arc",
    "arc",
    # errors
    "InvalidPath",
)

LEX_REX = re.compile(r"([MLHVCSQTAZmlhvcsqtaz])([^MLHVCSQTAZmlhvcsqtaz]*)")
NONE = lambda obj: obj is not None


class InvalidPath(ValueError):
    """Raised when given an invalid path string"""


class PathCommand(abc.ABC):
    """
    Base class of all path commands
    """

    # Number of arguments that follow this path commands letter
    nargs = -1

    @classproperty  # From python 3.9 on, just combine @classmethod and @property
    def name(cls):  # pylint: disable=no-self-argument
        """The full name of the segment (i.e. Line, Arc, etc)"""
        return cls.__name__  # pylint: disable=no-member

    @classproperty
    def letter(cls):  # pylint: disable=no-self-argument
        """The single letter representation of this command (i.e. L, A, etc)"""
        return cls.name[0]

    @classproperty
    def next_command(self):
        """The implicit next command. This is for automatic chains where the next
        command isn't given, just a bunch on numbers which we automatically parse."""
        return self

    @property
    def is_relative(self):  # type: () -> bool
        """Whether the command is defined in relative coordinates, i.e. relative to
        the previous endpoint (lower case path command letter)"""
        raise NotImplementedError

    @property
    def is_absolute(self):  # type: () -> bool
        """Whether the command is defined in absolute coordinates (upper case path
        command letter)"""
        raise NotImplementedError

    def to_relative(self, prev):  # type: (Vector2d) -> RelativePathCommand
        """Return absolute counterpart for absolute commands or copy for relative"""
        raise NotImplementedError

    def to_absolute(self, prev):  # type: (Vector2d) -> AbsolutePathCommand
        """Return relative counterpart for relative commands or copy for absolute"""
        raise NotImplementedError

    def reverse(self, first, prev):
        """Reverse path command

        .. versionadded:: 1.1"""

    def to_non_shorthand(self, prev, prev_control):  # pylint: disable=unused-argument
        # type: (Vector2d, Vector2d) -> AbsolutePathCommand
        """Return an absolute non-shorthand command

        .. versionadded:: 1.1"""
        return self.to_absolute(prev)

    # The precision of the numbers when converting to string
    number_template = "{:.6g}"

    # Maps single letter path command to corresponding class
    # (filled at the bottom of file, when all classes already defined)
    _letter_to_class = {}  # type: Dict[str, Type[Any]]

    @staticmethod
    def letter_to_class(letter):
        """Returns class for given path command letter"""
        return PathCommand._letter_to_class[letter]

    @property
    def args(self):  # type: () -> List[float]
        """Returns path command arguments as tuple of floats"""
        raise NotImplementedError()

    def control_points(
        self, first: Vector2d, prev: Vector2d, prev_prev: Vector2d
    ) -> Union[List[Vector2d], Generator[Vector2d, None, None]]:
        """Returns list of path command control points"""
        raise NotImplementedError

    @classmethod
    def _argt(cls, sep):
        return sep.join([cls.number_template] * cls.nargs)

    def __str__(self):
        return f"{self.letter} {self._argt(' ').format(*self.args)}".strip()

    def __repr__(self):
        # pylint: disable=consider-using-f-string
        return "{{}}({})".format(self._argt(", ")).format(self.name, *self.args)

    def __eq__(self, other):
        previous = Vector2d()
        if type(self) == type(other):  # pylint: disable=unidiomatic-typecheck
            return self.args == other.args
        if isinstance(other, tuple):
            return self.args == other
        if not isinstance(other, PathCommand):
            raise ValueError("Can't compare types")
        try:
            if self.is_relative == other.is_relative:
                return self.to_curve(previous) == other.to_curve(previous)
        except ValueError:
            pass
        return False

    def end_point(self, first, prev):  # type: (Vector2d, Vector2d) -> Vector2d
        """Returns last control point of path command"""
        raise NotImplementedError()

    def update_bounding_box(
        self, first: Vector2d, last_two_points: List[Vector2d], bbox: BoundingBox
    ):
        # pylint: disable=unused-argument
        """Enlarges given bbox to contain path element.

        Args:
            first (Vector2d): first point of path. Required to calculate Z segment
            last_two_points (List[Vector2d]): list with last two control points in abs
                coords.
            bbox (BoundingBox): bounding box to update
        """

        raise NotImplementedError(f"Bounding box is not implemented for {self.name}")

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> Curve
        """Convert command to :py:class:`Curve`

        Curve().to_curve() returns a copy
        """
        raise NotImplementedError(f"To curve not supported for {self.name}")

    def to_curves(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> List[Curve]
        """Convert command to list of :py:class:`Curve` commands"""
        return [self.to_curve(prev, prev_prev)]

    def to_line(self, prev):
        # type: (Vector2d) -> Line
        """Converts this segment to a line (copies if already a line)"""
        return Line(*self.end_point(Vector2d(), prev))


class RelativePathCommand(PathCommand):
    """
    Abstract base class for relative path commands.

    Implements most of methods of :py:class:`PathCommand` through
    conversion to :py:class:`AbsolutePathCommand`
    """

    @property
    def is_relative(self):
        return True

    @property
    def is_absolute(self):
        return False

    def control_points(
        self, first: Vector2d, prev: Vector2d, prev_prev: Vector2d
    ) -> Union[List[Vector2d], Generator[Vector2d, None, None]]:
        return self.to_absolute(prev).control_points(first, prev, prev_prev)

    def to_relative(self, prev):
        # type: (Pathlike, Vector2d) -> Pathlike
        return self.__class__(*self.args)

    def update_bounding_box(self, first, last_two_points, bbox):
        self.to_absolute(last_two_points[-1]).update_bounding_box(
            first, last_two_points, bbox
        )

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return self.to_absolute(prev).end_point(first, prev)

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> Curve
        return self.to_absolute(prev).to_curve(prev, prev_prev)

    def to_curves(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> List[Curve]
        return self.to_absolute(prev).to_curves(prev, prev_prev)


class AbsolutePathCommand(PathCommand):
    """Absolute path command. Unlike :py:class:`RelativePathCommand` can be transformed
    directly."""

    @property
    def is_relative(self):
        return False

    @property
    def is_absolute(self):
        return True

    def to_absolute(
        self, prev
    ):  # type: (AbsolutePathlike, Vector2d) -> AbsolutePathlike
        return self.__class__(*self.args)

    def transform(
        self, transform
    ):  # type: (AbsolutePathlike, Transform) -> AbsolutePathlike
        """Returns new transformed segment

        :param transform: a transformation to apply
        """
        raise NotImplementedError()

    def rotate(
        self, degrees, center
    ):  # type: (AbsolutePathlike, float, Vector2d) -> AbsolutePathlike
        """
        Returns new transformed segment

        :param degrees: rotation angle in degrees
        :param center: invariant point of rotation
        """
        return self.transform(Transform(rotate=(degrees, center[0], center[1])))

    def translate(self, dr):  # type: (AbsolutePathlike, Vector2d) -> AbsolutePathlike
        """Translate or scale this path command by dr"""
        return self.transform(Transform(translate=dr))

    def scale(
        self, factor
    ):  # type: (AbsolutePathlike, Union[float, Tuple[float,float]]) -> AbsolutePathlike
        """Returns new transformed segment

        :param factor: scale or (scale_x, scale_y)
        """
        return self.transform(Transform(scale=factor))


class Line(AbsolutePathCommand):
    """Line segment"""

    nargs = 2

    @property
    def args(self):
        return self.x, self.y

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox(
            (last_two_points[-1].x, self.x), (last_two_points[-1].y, self.y)
        )

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield Vector2d(self.x, self.y)

    def to_relative(self, prev):
        # type: (Vector2d) -> line
        return line(self.x - prev.x, self.y - prev.y)

    def transform(self, transform):
        # type: (Line, Transform) -> Line
        return Line(*transform.apply_to_point((self.x, self.y)))

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return Vector2d(self.x, self.y)

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Optional[Vector2d]) -> Curve
        return Curve(prev.x, prev.y, self.x, self.y, self.x, self.y)

    def reverse(self, first, prev):
        return Line(prev.x, prev.y)


class line(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative line segment"""

    nargs = 2

    @property
    def args(self):
        return self.dx, self.dy

    def __init__(self, dx, dy):
        self.dx = dx
        self.dy = dy

    def to_absolute(self, prev):  # type: (Vector2d) -> Line
        return Line(prev.x + self.dx, prev.y + self.dy)

    def reverse(self, first, prev):
        return line(-self.dx, -self.dy)


class Move(AbsolutePathCommand):
    """Move pen segment without a line"""

    nargs = 2
    next_command = Line

    @property
    def args(self):
        return self.x, self.y

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox(self.x, self.y)

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield Vector2d(self.x, self.y)

    def to_relative(self, prev):
        # type: (Vector2d) -> move
        return move(self.x - prev.x, self.y - prev.y)

    def transform(self, transform):
        # type: (Transform) -> Move
        return Move(*transform.apply_to_point((self.x, self.y)))

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return Vector2d(self.x, self.y)

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Optional[Vector2d]) -> Curve
        raise ValueError("Move segments can not be changed into curves.")

    def reverse(self, first, prev):
        return Move(prev.x, prev.y)


class move(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative move segment"""

    nargs = 2
    next_command = line

    @property
    def args(self):
        return self.dx, self.dy

    def __init__(self, dx, dy):
        self.dx = dx
        self.dy = dy

    def to_absolute(self, prev):  # type: (Vector2d) -> Move
        return Move(prev.x + self.dx, prev.y + self.dy)

    def reverse(self, first, prev):
        return move(prev.x - first.x, prev.y - first.y)


class ZoneClose(AbsolutePathCommand):
    """Close segment to finish a path"""

    nargs = 0
    next_command = Move

    @property
    def args(self):
        return ()

    def update_bounding_box(self, first, last_two_points, bbox):
        pass

    def transform(self, transform):
        # type: (Transform) -> ZoneClose
        return ZoneClose()

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield first

    def to_relative(self, prev):
        # type: (Vector2d) -> zoneClose
        return zoneClose()

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return first

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Optional[Vector2d]) -> Curve
        raise ValueError("ZoneClose segments can not be changed into curves.")

    def reverse(self, first, prev):
        return Line(prev.x, prev.y)


class zoneClose(RelativePathCommand):  # pylint: disable=invalid-name
    """Same as above (svg says no difference)"""

    nargs = 0
    next_command = Move

    @property
    def args(self):
        return ()

    def to_absolute(self, prev):
        return ZoneClose()

    def reverse(self, first, prev):
        return line(prev.x - first.x, prev.y - first.y)


class Horz(AbsolutePathCommand):
    """Horizontal Line segment"""

    nargs = 1

    @property
    def args(self):
        return (self.x,)

    def __init__(self, x):
        self.x = x

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox((last_two_points[-1].x, self.x), last_two_points[-1].y)

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield Vector2d(self.x, prev.y)

    def to_relative(self, prev):
        # type: (Vector2d) -> horz
        return horz(self.x - prev.x)

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> Line
        return self.to_line(prev)

    def transform(self, transform):
        # type: (Pathlike, Transform) -> Pathlike
        raise ValueError("Horizontal lines can't be transformed directly.")

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return Vector2d(self.x, prev.y)

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Optional[Vector2d]) -> Curve
        """Convert a horizontal line into a curve"""
        return self.to_line(prev).to_curve(prev)

    def to_line(self, prev):
        # type: (Vector2d) -> Line
        """Return this path command as a Line instead"""
        return Line(self.x, prev.y)

    def reverse(self, first, prev):
        return Horz(prev.x)


class horz(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative horz line segment"""

    nargs = 1

    @property
    def args(self):
        return (self.dx,)

    def __init__(self, dx):
        self.dx = dx

    def to_absolute(self, prev):  # type: (Vector2d) -> Horz
        return Horz(prev.x + self.dx)

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> Line
        return self.to_line(prev)

    def to_line(self, prev):  # type: (Vector2d) -> Line
        """Return this path command as a Line instead"""
        return Line(prev.x + self.dx, prev.y)

    def reverse(self, first, prev):
        return horz(-self.dx)


class Vert(AbsolutePathCommand):
    """Vertical Line segment"""

    nargs = 1

    @property
    def args(self):
        return (self.y,)

    def __init__(self, y):
        self.y = y

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox(last_two_points[-1].x, (last_two_points[-1].y, self.y))

    def transform(self, transform):  # type: (Pathlike, Transform) -> Pathlike
        raise ValueError("Vertical lines can't be transformed directly.")

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield Vector2d(prev.x, self.y)

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> Line
        return self.to_line(prev)

    def to_relative(self, prev):
        # type: (Vector2d) -> vert
        return vert(self.y - prev.y)

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return Vector2d(prev.x, self.y)

    def to_line(self, prev):
        # type: (Vector2d) -> Line
        """Return this path command as a line instead"""
        return Line(prev.x, self.y)

    def to_curve(
        self, prev, prev_prev=Vector2d()
    ):  # type: (Vector2d, Optional[Vector2d]) -> Curve
        """Convert a horizontal line into a curve"""
        return self.to_line(prev).to_curve(prev)

    def reverse(self, first, prev):
        return Vert(prev.y)


class vert(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative vertical line segment"""

    nargs = 1

    @property
    def args(self):
        return (self.dy,)

    def __init__(self, dy):
        self.dy = dy

    def to_absolute(self, prev):  # type: (Vector2d) -> Vert
        return Vert(prev.y + self.dy)

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> Line
        return self.to_line(prev)

    def to_line(self, prev):  # type: (Vector2d) -> Line
        """Return this path command as a line instead"""
        return Line(prev.x, prev.y + self.dy)

    def reverse(self, first, prev):
        return vert(-self.dy)


class Curve(AbsolutePathCommand):
    """Absolute Curved Line segment"""

    nargs = 6

    @property
    def args(self):
        return self.x2, self.y2, self.x3, self.y3, self.x4, self.y4

    def __init__(self, x2, y2, x3, y3, x4, y4):  # pylint: disable=too-many-arguments
        self.x2 = x2
        self.y2 = y2

        self.x3 = x3
        self.y3 = y3

        self.x4 = x4
        self.y4 = y4

    def update_bounding_box(self, first, last_two_points, bbox):

        x1, x2, x3, x4 = last_two_points[-1].x, self.x2, self.x3, self.x4
        y1, y2, y3, y4 = last_two_points[-1].y, self.y2, self.y3, self.y4

        if not (x1 in bbox.x and x2 in bbox.x and x3 in bbox.x and x4 in bbox.x):
            bbox.x += cubic_extrema(x1, x2, x3, x4)

        if not (y1 in bbox.y and y2 in bbox.y and y3 in bbox.y and y4 in bbox.y):
            bbox.y += cubic_extrema(y1, y2, y3, y4)

    def transform(self, transform):
        # type: (Transform) -> Curve
        x2, y2 = transform.apply_to_point((self.x2, self.y2))
        x3, y3 = transform.apply_to_point((self.x3, self.y3))
        x4, y4 = transform.apply_to_point((self.x4, self.y4))
        return Curve(x2, y2, x3, y3, x4, y4)

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield Vector2d(self.x2, self.y2)
        yield Vector2d(self.x3, self.y3)
        yield Vector2d(self.x4, self.y4)

    def to_relative(self, prev):  # type: (Vector2d) -> curve
        return curve(
            self.x2 - prev.x,
            self.y2 - prev.y,
            self.x3 - prev.x,
            self.y3 - prev.y,
            self.x4 - prev.x,
            self.y4 - prev.y,
        )

    def end_point(self, first, prev):
        return Vector2d(self.x4, self.y4)

    def to_curve(
        self, prev, prev_prev=Vector2d()
    ):  # type: (Vector2d, Optional[Vector2d]) -> Curve
        """No conversion needed, pass-through, returns self"""
        return Curve(*self.args)

    def to_bez(self):
        """Returns the list of coords for SuperPath"""
        return [list(self.args[:2]), list(self.args[2:4]), list(self.args[4:6])]

    def reverse(self, first, prev):
        return Curve(self.x3, self.y3, self.x2, self.y2, prev.x, prev.y)


class curve(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative curved line segment"""

    nargs = 6

    @property
    def args(self):
        return self.dx2, self.dy2, self.dx3, self.dy3, self.dx4, self.dy4

    def __init__(
        self, dx2, dy2, dx3, dy3, dx4, dy4
    ):  # pylint: disable=too-many-arguments
        self.dx2 = dx2
        self.dy2 = dy2

        self.dx3 = dx3
        self.dy3 = dy3

        self.dx4 = dx4
        self.dy4 = dy4

    def to_absolute(self, prev):  # type: (Vector2d) -> Curve
        return Curve(
            self.dx2 + prev.x,
            self.dy2 + prev.y,
            self.dx3 + prev.x,
            self.dy3 + prev.y,
            self.dx4 + prev.x,
            self.dy4 + prev.y,
        )

    def reverse(self, first, prev):
        return curve(
            -self.dx4 + self.dx3,
            -self.dy4 + self.dy3,
            -self.dx4 + self.dx2,
            -self.dy4 + self.dy2,
            -self.dx4,
            -self.dy4,
        )


class Smooth(AbsolutePathCommand):
    """Absolute Smoothed Curved Line segment"""

    nargs = 4

    @property
    def args(self):
        return self.x3, self.y3, self.x4, self.y4

    def __init__(self, x3, y3, x4, y4):

        self.x3 = x3
        self.y3 = y3

        self.x4 = x4
        self.y4 = y4

    def update_bounding_box(self, first, last_two_points, bbox):
        self.to_curve(last_two_points[-1], last_two_points[-2]).update_bounding_box(
            first, last_two_points, bbox
        )

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]

        x1, x2, x3, x4 = prev_prev.x, prev.x, self.x3, self.x4
        y1, y2, y3, y4 = prev_prev.y, prev.y, self.y3, self.y4

        # infer reflected point
        x2 = 2 * x2 - x1
        y2 = 2 * y2 - y1

        yield Vector2d(x2, y2)
        yield Vector2d(x3, y3)
        yield Vector2d(x4, y4)

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> Curve
        return self.to_curve(prev, prev_control)

    def to_relative(self, prev):  # type: (Vector2d) -> smooth
        return smooth(
            self.x3 - prev.x, self.y3 - prev.y, self.x4 - prev.x, self.y4 - prev.y
        )

    def transform(self, transform):
        # type: (Transform) -> Smooth
        x3, y3 = transform.apply_to_point((self.x3, self.y3))
        x4, y4 = transform.apply_to_point((self.x4, self.y4))
        return Smooth(x3, y3, x4, y4)

    def end_point(self, first, prev):
        return Vector2d(self.x4, self.y4)

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> Curve
        """
        Convert this Smooth curve to a regular curve by creating a mirror
        set of nodes based on the previous node. Previous should be a curve.
        """
        (x2, y2), (x3, y3), (x4, y4) = self.control_points(prev, prev, prev_prev)
        return Curve(x2, y2, x3, y3, x4, y4)

    def reverse(self, first, prev):
        return Smooth(self.x3, self.y3, prev.x, prev.y)


class smooth(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative smoothed curved line segment"""

    nargs = 4

    @property
    def args(self):
        return self.dx3, self.dy3, self.dx4, self.dy4

    def __init__(self, dx3, dy3, dx4, dy4):
        self.dx3 = dx3
        self.dy3 = dy3

        self.dx4 = dx4
        self.dy4 = dy4

    def to_absolute(self, prev):  # type: (Vector2d) -> Smooth
        return Smooth(
            self.dx3 + prev.x, self.dy3 + prev.y, self.dx4 + prev.x, self.dy4 + prev.y
        )

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> Curve
        return self.to_absolute(prev).to_non_shorthand(prev, prev_control)

    def reverse(self, first, prev):
        return smooth(-self.dx4 + self.dx3, -self.dy4 + self.dy3, -self.dx4, -self.dy4)


class Quadratic(AbsolutePathCommand):
    """Absolute Quadratic Curved Line segment"""

    nargs = 4

    @property
    def args(self):
        return self.x2, self.y2, self.x3, self.y3

    def __init__(self, x2, y2, x3, y3):

        self.x2 = x2
        self.y2 = y2

        self.x3 = x3
        self.y3 = y3

    def update_bounding_box(self, first, last_two_points, bbox):

        x1, x2, x3 = last_two_points[-1].x, self.x2, self.x3
        y1, y2, y3 = last_two_points[-1].y, self.y2, self.y3

        if not (x1 in bbox.x and x2 in bbox.x and x3 in bbox.x):
            bbox.x += quadratic_extrema(x1, x2, x3)

        if not (y1 in bbox.y and y2 in bbox.y and y3 in bbox.y):
            bbox.y += quadratic_extrema(y1, y2, y3)

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield Vector2d(self.x2, self.y2)
        yield Vector2d(self.x3, self.y3)

    def to_relative(self, prev):
        # type: (Vector2d) -> quadratic
        return quadratic(
            self.x2 - prev.x, self.y2 - prev.y, self.x3 - prev.x, self.y3 - prev.y
        )

    def transform(self, transform):
        # type: (Transform) -> Quadratic
        x2, y2 = transform.apply_to_point((self.x2, self.y2))
        x3, y3 = transform.apply_to_point((self.x3, self.y3))
        return Quadratic(x2, y2, x3, y3)

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return Vector2d(self.x3, self.y3)

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> Curve
        """Attempt to convert a quadratic to a curve"""
        prev = Vector2d(prev)
        x1 = 1.0 / 3 * prev.x + 2.0 / 3 * self.x2
        x2 = 2.0 / 3 * self.x2 + 1.0 / 3 * self.x3
        y1 = 1.0 / 3 * prev.y + 2.0 / 3 * self.y2
        y2 = 2.0 / 3 * self.y2 + 1.0 / 3 * self.y3
        return Curve(x1, y1, x2, y2, self.x3, self.y3)

    def reverse(self, first, prev):
        return Quadratic(self.x2, self.y2, prev.x, prev.y)


class quadratic(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative quadratic line segment"""

    nargs = 4

    @property
    def args(self):
        return self.dx2, self.dy2, self.dx3, self.dy3

    def __init__(self, dx2, dy2, dx3, dy3):
        self.dx2 = dx2
        self.dx3 = dx3
        self.dy2 = dy2
        self.dy3 = dy3

    def to_absolute(self, prev):  # type: (Vector2d) -> Quadratic
        return Quadratic(
            self.dx2 + prev.x, self.dy2 + prev.y, self.dx3 + prev.x, self.dy3 + prev.y
        )

    def reverse(self, first, prev):
        return quadratic(
            -self.dx3 + self.dx2, -self.dy3 + self.dy2, -self.dx3, -self.dy3
        )


class TepidQuadratic(AbsolutePathCommand):
    """Continued Quadratic Line segment"""

    nargs = 2

    @property
    def args(self):
        return self.x3, self.y3

    def __init__(self, x3, y3):
        self.x3 = x3
        self.y3 = y3

    def update_bounding_box(self, first, last_two_points, bbox):
        self.to_quadratic(last_two_points[-1], last_two_points[-2]).update_bounding_box(
            first, last_two_points, bbox
        )

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]

        x1, x2, x3 = prev_prev.x, prev.x, self.x3
        y1, y2, y3 = prev_prev.y, prev.y, self.y3

        # infer reflected point
        x2 = 2 * x2 - x1
        y2 = 2 * y2 - y1

        yield Vector2d(x2, y2)
        yield Vector2d(x3, y3)

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> AbsolutePathCommand
        return self.to_quadratic(prev, prev_control)

    def to_relative(self, prev):  # type: (Vector2d) -> tepidQuadratic
        return tepidQuadratic(self.x3 - prev.x, self.y3 - prev.y)

    def transform(self, transform):
        # type: (Transform) -> TepidQuadratic
        x3, y3 = transform.apply_to_point((self.x3, self.y3))
        return TepidQuadratic(x3, y3)

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return Vector2d(self.x3, self.y3)

    def to_curve(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> Curve
        return self.to_quadratic(prev, prev_prev).to_curve(prev)

    def to_quadratic(self, prev, prev_prev):
        # type: (Vector2d, Vector2d) -> Quadratic
        """
        Convert this continued quadratic into a full quadratic
        """
        (x2, y2), (x3, y3) = self.control_points(prev, prev, prev_prev)
        return Quadratic(x2, y2, x3, y3)

    def reverse(self, first, prev):
        return TepidQuadratic(prev.x, prev.y)


class tepidQuadratic(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative continued quadratic line segment"""

    nargs = 2

    @property
    def args(self):
        return self.dx3, self.dy3

    def __init__(self, dx3, dy3):
        self.dx3 = dx3
        self.dy3 = dy3

    def to_absolute(self, prev):
        # type: (Vector2d) -> TepidQuadratic
        return TepidQuadratic(self.dx3 + prev.x, self.dy3 + prev.y)

    def to_non_shorthand(self, prev, prev_control):
        # type: (Vector2d, Vector2d) -> AbsolutePathCommand
        return self.to_absolute(prev).to_non_shorthand(prev, prev_control)

    def reverse(self, first, prev):
        return tepidQuadratic(-self.dx3, -self.dy3)


class Arc(AbsolutePathCommand):
    """Special Arc segment"""

    nargs = 7

    @property
    def args(self):
        return (
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.x,
            self.y,
        )

    def __init__(
        self, rx, ry, x_axis_rotation, large_arc, sweep, x, y
    ):  # pylint: disable=too-many-arguments
        self.rx = rx
        self.ry = ry
        self.x_axis_rotation = x_axis_rotation
        self.large_arc = large_arc
        self.sweep = sweep
        self.x = x
        self.y = y

    def update_bounding_box(self, first, last_two_points, bbox):
        prev = last_two_points[-1]
        for seg in self.to_curves(prev=prev):
            seg.update_bounding_box(first, [None, prev], bbox)
            prev = seg.end_point(first, prev)

    def control_points(self, first, prev, prev_prev):
        # type: (Vector2d, Vector2d, Vector2d) -> Generator[Vector2d, None, None]
        yield Vector2d(self.x, self.y)

    def to_curves(self, prev, prev_prev=Vector2d()):
        # type: (Vector2d, Vector2d) -> List[Curve]
        """Convert this arc into bezier curves"""
        path = CubicSuperPath([arc_to_path(list(prev), self.args)]).to_path(
            curves_only=True
        )
        # Ignore the first move command from to_path()
        return list(path)[1:]

    def transform(self, transform):
        # type: (Transform) -> Arc
        # pylint: disable=invalid-name, too-many-locals
        x_, y_ = transform.apply_to_point((self.x, self.y))

        T = transform  # type: Transform
        if self.x_axis_rotation != 0:
            T = T @ Transform(rotate=self.x_axis_rotation)
        a, c, b, d, _, _ = list(T.to_hexad())
        # T = | a b |
        #     | c d |

        detT = a * d - b * c
        detT2 = detT**2

        rx = float(self.rx)
        ry = float(self.ry)

        if rx == 0.0 or ry == 0.0 or detT2 == 0.0:
            # invalid Arc parameters
            # transform only last point
            return Arc(
                self.rx,
                self.ry,
                self.x_axis_rotation,
                self.large_arc,
                self.sweep,
                x_,
                y_,
            )

        A = (d**2 / rx**2 + c**2 / ry**2) / detT2
        B = -(d * b / rx**2 + c * a / ry**2) / detT2
        D = (b**2 / rx**2 + a**2 / ry**2) / detT2

        theta = atan2(-2 * B, D - A) / 2
        theta_deg = theta * 180.0 / pi
        DA = D - A
        l2 = 4 * B**2 + DA**2

        if l2 == 0:
            delta = 0.0
        else:
            delta = 0.5 * (-(DA**2) - 4 * B**2) / sqrt(l2)

        half = (A + D) / 2

        rx_ = 1.0 / sqrt(half + delta)
        ry_ = 1.0 / sqrt(half - delta)

        x_, y_ = transform.apply_to_point((self.x, self.y))

        if detT > 0:
            sweep = self.sweep
        else:
            sweep = 0 if self.sweep > 0 else 1

        return Arc(rx_, ry_, theta_deg, self.large_arc, sweep, x_, y_)

    def to_relative(self, prev):
        # type: (Vector2d) -> arc
        return arc(
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.x - prev.x,
            self.y - prev.y,
        )

    def end_point(self, first, prev):
        # type: (Vector2d, Vector2d) -> Vector2d
        return Vector2d(self.x, self.y)

    def reverse(self, first, prev):
        return Arc(
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            1 - self.sweep,
            prev.x,
            prev.y,
        )


class arc(RelativePathCommand):  # pylint: disable=invalid-name
    """Relative Arc line segment"""

    nargs = 7

    @property
    def args(self):
        return (
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.dx,
            self.dy,
        )

    def __init__(
        self, rx, ry, x_axis_rotation, large_arc, sweep, dx, dy
    ):  # pylint: disable=too-many-arguments
        self.rx = rx
        self.ry = ry
        self.x_axis_rotation = x_axis_rotation
        self.large_arc = large_arc
        self.sweep = sweep
        self.dx = dx
        self.dy = dy

    def to_absolute(self, prev):  # type: (Vector2d) -> "Arc"
        x1, y1 = prev
        return Arc(
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            self.sweep,
            self.dx + x1,
            self.dy + y1,
        )

    def reverse(self, first, prev):
        return arc(
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.large_arc,
            1 - self.sweep,
            -self.dx,
            -self.dy,
        )


PathCommand._letter_to_class = {  # pylint: disable=protected-access
    "M": Move,
    "L": Line,
    "V": Vert,
    "H": Horz,
    "A": Arc,
    "C": Curve,
    "S": Smooth,
    "Z": ZoneClose,
    "Q": Quadratic,
    "T": TepidQuadratic,
    "m": move,
    "l": line,
    "v": vert,
    "h": horz,
    "a": arc,
    "c": curve,
    "s": smooth,
    "z": zoneClose,
    "q": quadratic,
    "t": tepidQuadratic,
}


class Path(list):
    """A list of segment commands which combine to draw a shape"""

    class PathCommandProxy:
        """
        A handy class for Path traverse and coordinate access

        Reduces number of arguments in user code compared to bare
        :class:`PathCommand` methods
        """

        def __init__(
            self, command, first_point, previous_end_point, prev2_control_point
        ):
            self.command = command  # type: PathCommand
            self.first_point = first_point  # type: Vector2d
            self.previous_end_point = previous_end_point  # type: Vector2d
            self.prev2_control_point = prev2_control_point  # type: Vector2d

        @property
        def name(self):
            """The full name of the segment (i.e. Line, Arc, etc)"""
            return self.command.name

        @property
        def letter(self):
            """The single letter representation of this command (i.e. L, A, etc)"""
            return self.command.letter

        @property
        def next_command(self):
            """The implicit next command."""
            return self.command.next_command

        @property
        def is_relative(self):
            """Whether the command is defined in relative coordinates, i.e. relative to
            the previous endpoint (lower case path command letter)"""
            return self.command.is_relative

        @property
        def is_absolute(self):
            """Whether the command is defined in absolute coordinates (upper case path
            command letter)"""
            return self.command.is_absolute

        @property
        def args(self):
            """Returns path command arguments as tuple of floats"""
            return self.command.args

        @property
        def control_points(self):
            """Returns list of path command control points"""
            return self.command.control_points(
                self.first_point, self.previous_end_point, self.prev2_control_point
            )

        @property
        def end_point(self):
            """Returns last control point of path command"""
            return self.command.end_point(self.first_point, self.previous_end_point)

        def reverse(self):
            """Reverse path command"""
            return self.command.reverse(self.end_point, self.previous_end_point)

        def to_curve(self):
            """Convert command to :py:class:`Curve`
            Curve().to_curve() returns a copy
            """
            return self.command.to_curve(
                self.previous_end_point, self.prev2_control_point
            )

        def to_curves(self):
            """Convert command to list of :py:class:`Curve` commands"""
            return self.command.to_curves(
                self.previous_end_point, self.prev2_control_point
            )

        def to_absolute(self):
            """Return relative counterpart for relative commands or copy for absolute"""
            return self.command.to_absolute(self.previous_end_point)

        def __str__(self):
            return str(self.command)

        def __repr__(self):
            return "<" + self.__class__.__name__ + ">" + repr(self.command)

    def __init__(self, path_d=None):
        super().__init__()
        if isinstance(path_d, str):
            # Returns a generator returning PathCommand objects
            path_d = self.parse_string(path_d)
        elif isinstance(path_d, CubicSuperPath):
            path_d = path_d.to_path()

        for item in path_d or ():
            if isinstance(item, PathCommand):
                self.append(item)
            elif isinstance(item, (list, tuple)) and len(item) == 2:
                if isinstance(item[1], (list, tuple)):
                    self.append(PathCommand.letter_to_class(item[0])(*item[1]))
                else:
                    self.append(Line(*item))
            else:
                raise TypeError(
                    f"Bad path type: {type(path_d).__name__}"
                    f"({type(item).__name__}, ...): {item}"
                )

    @classmethod
    def parse_string(cls, path_d):
        """Parse a path string and generate segment objects"""
        for cmd, numbers in LEX_REX.findall(path_d):
            args = list(strargs(numbers))
            cmd = PathCommand.letter_to_class(cmd)
            i = 0
            while i < len(args) or cmd.nargs == 0:
                if len(args[i : i + cmd.nargs]) != cmd.nargs:
                    return
                seg = cmd(*args[i : i + cmd.nargs])
                i += cmd.nargs
                cmd = seg.next_command
                yield seg

    def bounding_box(self):
        # type: () -> Optional[BoundingBox]
        """Return bounding box of the Path"""
        if not self:
            return None
        iterator = self.proxy_iterator()
        proxy = next(iterator)
        bbox = BoundingBox(proxy.first_point.x, proxy.first_point.y)
        try:
            while True:
                proxy = next(iterator)
                proxy.command.update_bounding_box(
                    proxy.first_point,
                    [
                        proxy.prev2_control_point,
                        proxy.previous_end_point,
                    ],
                    bbox,
                )
        except StopIteration:
            return bbox

    def append(self, cmd):
        """Append a command to this path including any chained commands"""
        if isinstance(cmd, list):
            self.extend(cmd)
        elif isinstance(cmd, PathCommand):
            super().append(cmd)

    def translate(self, x, y, inplace=False):  # pylint: disable=invalid-name
        """Move all coords in this path by the given amount"""
        return self.transform(Transform(translate=(x, y)), inplace=inplace)

    def scale(self, x, y, inplace=False):  # pylint: disable=invalid-name
        """Scale all coords in this path by the given amounts"""
        return self.transform(Transform(scale=(x, y)), inplace=inplace)

    def rotate(self, deg, center=None, inplace=False):
        """Rotate the path around the given point"""
        if center is None:
            # Default center is center of bbox
            bbox = self.bounding_box()
            if bbox:
                center = bbox.center
            else:
                center = Vector2d()
        center = Vector2d(center)
        return self.transform(
            Transform(rotate=(deg, center.x, center.y)), inplace=inplace
        )

    @property
    def control_points(self):
        """Returns all control points of the Path"""
        prev = Vector2d()
        prev_prev = Vector2d()
        first = Vector2d()

        for seg in self:  # type: PathCommand
            cpts = list(seg.control_points(first, prev, prev_prev))
            if isinstance(seg, (zoneClose, ZoneClose, move, Move)):
                first = cpts[-1]
            for cpt in cpts:
                prev_prev = prev
                prev = cpt
                yield cpt

    @property
    def end_points(self):
        """Returns all endpoints of all path commands (i.e. the nodes)"""
        prev = Vector2d()
        first = Vector2d()

        for seg in self:  # type: PathCommand
            end_point = seg.end_point(first, prev)
            if isinstance(seg, (zoneClose, ZoneClose, move, Move)):
                first = end_point
            prev = end_point
            yield end_point

    def transform(self, transform, inplace=False):
        """Convert to new path"""
        result = Path()
        previous = Vector2d()
        previous_new = Vector2d()
        start_zone = True
        first = Vector2d()
        first_new = Vector2d()

        for i, seg in enumerate(self):  # type: PathCommand
            if start_zone:
                first = seg.end_point(first, previous)

            if isinstance(seg, (horz, Horz, Vert, vert)):
                seg = seg.to_line(previous)

            if seg.is_relative:
                new_seg = (
                    seg.to_absolute(previous)
                    .transform(transform)
                    .to_relative(previous_new)
                )
            else:
                new_seg = seg.transform(transform)

            if start_zone:
                first_new = new_seg.end_point(first_new, previous_new)

            if inplace:
                self[i] = new_seg
            else:
                result.append(new_seg)
            previous = seg.end_point(first, previous)
            previous_new = new_seg.end_point(first_new, previous_new)
            start_zone = isinstance(seg, (zoneClose, ZoneClose))
        if inplace:
            return self
        return result

    def reverse(self):
        """Returns a reversed path"""
        result = Path()
        *_, first = self.end_points
        closer = None

        # Go through the path in reverse order
        for index, prcom in reversed(list(enumerate(self.proxy_iterator()))):
            if isinstance(prcom.command, (Move, move, ZoneClose, zoneClose)):
                if closer is not None:
                    if len(result) > 0 and isinstance(
                        result[-1], (Line, line, Vert, vert, Horz, horz)
                    ):
                        result.pop()  # We can replace simple lines with Z
                    result.append(closer)  # replace with same type (rel or abs)
                if isinstance(prcom.command, (ZoneClose, zoneClose)):
                    closer = prcom.command
                else:
                    closer = None

            if index == 0:
                if prcom.letter == "M":
                    result.insert(0, Move(first.x, first.y))
                elif prcom.letter == "m":
                    result.insert(0, move(first.x, first.y))
            else:
                result.append(prcom.reverse())

        return result

    def close(self):
        """Attempt to close the last path segment"""
        if self and not isinstance(self[-1], (zoneClose, ZoneClose)):
            self.append(ZoneClose())

    def proxy_iterator(self):
        """
        Yields :py:class:`AugmentedPathIterator`

        :rtype: Iterator[ Path.PathCommandProxy ]
        """

        previous = Vector2d()
        prev_prev = Vector2d()
        first = Vector2d()

        for seg in self:  # type: PathCommand#
            if isinstance(seg, (zoneClose, ZoneClose, move, Move)):
                first = seg.end_point(first, previous)
            yield Path.PathCommandProxy(seg, first, previous, prev_prev)
            if isinstance(
                seg,
                (
                    curve,
                    tepidQuadratic,
                    quadratic,
                    smooth,
                    Curve,
                    TepidQuadratic,
                    Quadratic,
                    Smooth,
                ),
            ):
                prev_prev = list(seg.control_points(first, previous, prev_prev))[-2]
            previous = seg.end_point(first, previous)

    def to_absolute(self):
        """Convert this path to use only absolute coordinates"""
        return self._to_absolute(True)

    def to_non_shorthand(self):
        # type: () -> Path
        """Convert this path to use only absolute non-shorthand coordinates

        .. versionadded:: 1.1"""
        return self._to_absolute(False)

    def _to_absolute(self, shorthand: bool) -> Path:
        """Make entire Path absolute.

        Args:
            shorthand (bool): If false, then convert all shorthand commands to
                non-shorthand.

        Returns:
            Path: the input path, converted to absolute coordinates.
        """

        abspath = Path()

        previous = Vector2d()
        first = Vector2d()

        for seg in self:  # type: PathCommand
            if isinstance(seg, (move, Move)):
                first = seg.end_point(first, previous)

            if shorthand:
                abspath.append(seg.to_absolute(previous))
            else:
                if abspath and isinstance(abspath[-1], (Curve, Quadratic)):
                    prev_control = list(
                        abspath[-1].control_points(Vector2d(), Vector2d(), Vector2d())
                    )[-2]
                else:
                    prev_control = previous

                abspath.append(seg.to_non_shorthand(previous, prev_control))

            previous = seg.end_point(first, previous)

        return abspath

    def to_relative(self):
        """Convert this path to use only relative coordinates"""
        abspath = Path()

        previous = Vector2d()
        first = Vector2d()

        for seg in self:  # type: PathCommand
            if isinstance(seg, (move, Move)):
                first = seg.end_point(first, previous)

            abspath.append(seg.to_relative(previous))
            previous = seg.end_point(first, previous)

        return abspath

    def __str__(self):
        return " ".join([str(seg) for seg in self])

    def __add__(self, other):
        acopy = copy.deepcopy(self)
        if isinstance(other, str):
            other = Path(other)
        if isinstance(other, list):
            acopy.extend(other)
        return acopy

    def to_arrays(self):
        """Returns path in format of parsePath output, returning arrays of absolute
        command data

        .. deprecated:: 1.0
            This is compatibility function for older API. Should not be used in new code

        """
        return [[seg.letter, list(seg.args)] for seg in self.to_non_shorthand()]

    def to_superpath(self):
        """Convert this path into a cubic super path"""
        return CubicSuperPath(self)

    def copy(self):
        """Make a copy"""
        return copy.deepcopy(self)


class CubicSuperPath(list):
    """
    A conversion of a path into a predictable list of cubic curves which
    can be operated on as a list of simplified instructions.

    When converting back into a path, all lines, arcs etc will be converted
    to curve instructions.

    Structure is held as [SubPath[(point_a, bezier, point_b), ...]], ...]
    """

    def __init__(self, items):
        super().__init__()
        self._closed = True
        self._prev = Vector2d()
        self._prev_prev = Vector2d()

        if isinstance(items, str):
            items = Path(items)

        if isinstance(items, Path):
            items = items.to_absolute()

        for item in items:
            self.append(item)

    def __str__(self):
        return str(self.to_path())

    def append(self, item, force_shift=False):
        """Accept multiple different formats for the data

        .. versionchanged:: 1.2
            ``force_shift`` parameter has been added
        """
        if isinstance(item, list) and len(item) == 2 and isinstance(item[0], str):
            item = PathCommand.letter_to_class(item[0])(*item[1])
        coordinate_shift = True
        if isinstance(item, list) and len(item) == 3 and not force_shift:
            coordinate_shift = False
        is_quadratic = False
        if isinstance(item, PathCommand):
            if isinstance(item, Move):
                if self._closed is False:
                    super().append([])
                item = [list(item.args), list(item.args), list(item.args)]
            elif isinstance(item, ZoneClose) and self and self[-1]:
                # This duplicates the first segment to 'close' the path, it's appended
                # directly because we don't want to last coord to change for the final
                # segment.
                self[-1].append(
                    [self[-1][0][0][:], self[-1][0][1][:], self[-1][0][2][:]]
                )
                # Then adds a new subpath for the next shape (if any)
                self._closed = True
                self._prev.assign(self._first)
                return
            elif isinstance(item, Arc):
                # Arcs are made up of three curves (approximated)
                for arc_curve in item.to_curves(self._prev, self._prev_prev):
                    x2, y2, x3, y3, x4, y4 = arc_curve.args
                    self.append([[x2, y2], [x3, y3], [x4, y4]], force_shift=True)
                    self._prev_prev.assign(x3, y3)
                return
            else:
                is_quadratic = isinstance(
                    item, (Quadratic, TepidQuadratic, quadratic, tepidQuadratic)
                )
                if isinstance(item, (Horz, Vert)):
                    item = item.to_line(self._prev)
                prp = self._prev_prev
                if is_quadratic:
                    self._prev_prev = list(
                        item.control_points(self._first, self._prev, prp)
                    )[-2:-1][0]
                item = item.to_curve(self._prev, prp)

        if isinstance(item, Curve):
            # Curves are cut into three tuples for the super path.
            item = item.to_bez()

        if not isinstance(item, list):
            raise ValueError(f"Unknown super curve item type: {item}")

        if len(item) != 3 or not all(len(bit) == 2 for bit in item):
            # The item is already a subpath (usually from some other process)
            if len(item[0]) == 3 and all(len(bit) == 2 for bit in item[0]):
                super().append(self._clean(item))
                self._prev_prev = Vector2d(self[-1][-1][0])
                self._prev = Vector2d(self[-1][-1][1])
                return
            raise ValueError(f"Unknown super curve list format: {item}")

        if self._closed:
            # Closed means that the previous segment is closed so we need a new one
            # We always append to the last open segment. CSP starts out closed.
            self._closed = False
            super().append([])

        if coordinate_shift:
            if self[-1]:
                # The last tuple is replaced, it's the coords of where the next segment
                # will land.
                self[-1][-1][-1] = item[0][:]
            # The last coord is duplicated, but is expected to be replaced
            self[-1].append(item[1:] + copy.deepcopy(item)[-1:])
        else:
            # Item is already a csp segment and has already been shifted.
            self[-1].append(copy.deepcopy(item))

        self._prev = Vector2d(self[-1][-1][1])
        if not is_quadratic:
            self._prev_prev = Vector2d(self[-1][-1][0])

    def _clean(self, lst):
        """Recursively clean lists so they have the same type"""
        if isinstance(lst, (tuple, list)):
            return [self._clean(child) for child in lst]
        return lst

    @property
    def _first(self):
        try:
            return Vector2d(self[-1][0][0])
        except IndexError:
            return Vector2d()

    def to_path(self, curves_only=False, rtol=1e-5, atol=1e-8):
        """Convert the super path back to an svg path

        Arguments: see :func:`to_segments` for parameters"""
        return Path(list(self.to_segments(curves_only, rtol, atol)))

    def to_segments(self, curves_only=False, rtol=1e-5, atol=1e-8):
        """Generate a set of segments for this cubic super path

        Arguments:
            curves_only (bool, optional): If False, curves that can be represented
                by Lineto / ZoneClose commands, will be. Defaults to False.
            rtol (float, optional): relative tolerance, passed to :func:`is_line` and
                :func:`inkex.transforms.ImmutableVector2d.is_close` for checking if a
                line can be replaced by a ZoneClose command. Defaults to 1e-5.

                .. versionadded:: 1.2
            atol: absolute tolerance, passed to :func:`is_line` and
                :func:`inkex.transforms.ImmutableVector2d.is_close`. Defaults to 1e-8.

                .. versionadded:: 1.2"""
        for subpath in self:
            previous = []
            for segment in subpath:
                if not previous:
                    yield Move(*segment[1][:])
                elif self.is_line(previous, segment, rtol, atol) and not curves_only:
                    if segment is subpath[-1] and Vector2d(segment[1]).is_close(
                        subpath[0][1], rtol, atol
                    ):
                        yield ZoneClose()
                    else:
                        yield Line(*segment[1][:])
                else:
                    yield Curve(*(previous[2][:] + segment[0][:] + segment[1][:]))
                previous = segment

    def transform(self, transform):
        """Apply a transformation matrix to this super path"""
        return self.to_path().transform(transform).to_superpath()

    @staticmethod
    def is_on(pt_a, pt_b, pt_c, tol=1e-8):
        """Checks if point pt_a is on the line between points pt_b and pt_c

        .. versionadded:: 1.2"""
        return CubicSuperPath.collinear(pt_a, pt_b, pt_c, tol) and (
            CubicSuperPath.within(pt_a[0], pt_b[0], pt_c[0])
            if pt_a[0] != pt_b[0]
            else CubicSuperPath.within(pt_a[1], pt_b[1], pt_c[1])
        )

    @staticmethod
    def collinear(pt_a, pt_b, pt_c, tol=1e-8):
        """Checks if points pt_a, pt_b, pt_c lie on the same line,
        i.e. that the cross product (b-a) x (c-a) < tol

        .. versionadded:: 1.2"""
        return (
            abs(
                (pt_b[0] - pt_a[0]) * (pt_c[1] - pt_a[1])
                - (pt_c[0] - pt_a[0]) * (pt_b[1] - pt_a[1])
            )
            < tol
        )

    @staticmethod
    def within(val_b, val_a, val_c):
        """Checks if float val_b is between val_a and val_c

        .. versionadded:: 1.2"""
        return val_a <= val_b <= val_c or val_c <= val_b <= val_a

    @staticmethod
    def is_line(previous, segment, rtol=1e-5, atol=1e-8):
        """Check whether csp segment (two points) can be expressed as a line has retracted handles or the handles
        can be retracted without loss of information (i.e. both handles lie on the
        line)

        .. versionchanged:: 1.2
            Previously, it was only checked if both control points have retracted
            handles. Now it is also checked if the handles can be retracted without
            (visible) loss of information (i.e. both handles lie on the line connecting
            the nodes).

        Arguments:
            previous: first node in superpath notation
            segment: second node in superpath notation
            rtol (float, optional): relative tolerance, passed to
                :func:`inkex.transforms.ImmutableVector2d.is_close` for checking handle
                retraction. Defaults to 1e-5.

                .. versionadded:: 1.2
            atol (float, optional): absolute tolerance, passed to
                :func:`inkex.transforms.ImmutableVector2d.is_close` for checking handle
                retraction and
                :func:`inkex.paths.CubicSuperPath.is_on` for checking if all points
                (nodes + handles) lie on a line. Defaults to 1e-8.

                .. versionadded:: 1.2
        """

        retracted = Vector2d(previous[1]).is_close(
            previous[2], rtol, atol
        ) and Vector2d(segment[0]).is_close(segment[1], rtol, atol)

        if retracted:
            return True

        # Can both handles be retracted without loss of information?
        # Definitely the case if the handles lie on the same line as the two nodes and
        # in the correct order
        # E.g. cspbezsplitatlength outputs non-retracted handles when splitting a
        # straight line
        return CubicSuperPath.is_on(
            segment[0], segment[1], previous[2], atol
        ) and CubicSuperPath.is_on(previous[2], previous[1], segment[0], atol)


def arc_to_path(point, params):
    """Approximates an arc with cubic bezier segments.

    Arguments:
        point:  Starting point (absolute coords)
        params: Arcs parameters as per
              https://www.w3.org/TR/SVG/paths.html#PathDataEllipticalArcCommands

    Returns a list of triplets of points :
    [control_point_before, node, control_point_after]
    (first and last returned triplets are [p1, p1, *] and [*, p2, p2])
    """

    # pylint: disable=invalid-name, too-many-locals
    A = point[:]
    rx, ry, teta, longflag, sweepflag, x2, y2 = params[:]
    teta = teta * pi / 180.0
    B = [x2, y2]
    # Degenerate ellipse
    if rx == 0 or ry == 0 or A == B:
        return [[A[:], A[:], A[:]], [B[:], B[:], B[:]]]

    # turn coordinates so that the ellipse morph into a *unit circle* (not 0-centered)
    mat = matprod((rotmat(teta), [[1.0 / rx, 0.0], [0.0, 1.0 / ry]], rotmat(-teta)))
    applymat(mat, A)
    applymat(mat, B)

    k = [-(B[1] - A[1]), B[0] - A[0]]
    d = k[0] * k[0] + k[1] * k[1]
    k[0] /= sqrt(d)
    k[1] /= sqrt(d)
    d = sqrt(max(0, 1 - d / 4.0))
    # k is the unit normal to AB vector, pointing to center O
    # d is distance from center to AB segment (distance from O to the midpoint of AB)
    # for the last line, remember this is a unit circle, and kd vector is ortogonal to
    # AB (Pythagorean thm)

    if longflag == sweepflag:
        # top-right ellipse in SVG example
        # https://www.w3.org/TR/SVG/images/paths/arcs02.svg
        d *= -1

    O = [(B[0] + A[0]) / 2.0 + d * k[0], (B[1] + A[1]) / 2.0 + d * k[1]]
    OA = [A[0] - O[0], A[1] - O[1]]
    OB = [B[0] - O[0], B[1] - O[1]]
    start = acos(OA[0] / norm(OA))
    if OA[1] < 0:
        start *= -1
    end = acos(OB[0] / norm(OB))
    if OB[1] < 0:
        end *= -1
    # start and end are the angles from center of the circle to A and to B respectively

    if sweepflag and start > end:
        end += 2 * pi
    if (not sweepflag) and start < end:
        end -= 2 * pi

    NbSectors = int(abs(start - end) * 2 / pi) + 1
    dTeta = (end - start) / NbSectors
    v = 4 * tan(dTeta / 4.0) / 3.0
    # I would use v = tan(dTeta/2)*4*(sqrt(2)-1)/3 ?
    p = []
    for i in range(0, NbSectors + 1, 1):
        angle = start + i * dTeta
        v1 = [
            O[0] + cos(angle) - (-v) * sin(angle),
            O[1] + sin(angle) + (-v) * cos(angle),
        ]
        pt = [O[0] + cos(angle), O[1] + sin(angle)]
        v2 = [O[0] + cos(angle) - v * sin(angle), O[1] + sin(angle) + v * cos(angle)]
        p.append([v1, pt, v2])
    p[0][0] = p[0][1][:]
    p[-1][2] = p[-1][1][:]

    # go back to the original coordinate system
    mat = matprod((rotmat(teta), [[rx, 0], [0, ry]], rotmat(-teta)))
    for pts in p:
        applymat(mat, pts[0])
        applymat(mat, pts[1])
        applymat(mat, pts[2])
    return p


def matprod(mlist):
    """Get the product of the mat"""
    prod = mlist[0]
    for mat in mlist[1:]:
        a00 = prod[0][0] * mat[0][0] + prod[0][1] * mat[1][0]
        a01 = prod[0][0] * mat[0][1] + prod[0][1] * mat[1][1]
        a10 = prod[1][0] * mat[0][0] + prod[1][1] * mat[1][0]
        a11 = prod[1][0] * mat[0][1] + prod[1][1] * mat[1][1]
        prod = [[a00, a01], [a10, a11]]
    return prod


def rotmat(teta):
    """Rotate the mat"""
    return [[cos(teta), -sin(teta)], [sin(teta), cos(teta)]]


def applymat(mat, point):
    """Apply the given mat"""
    x = mat[0][0] * point[0] + mat[0][1] * point[1]
    y = mat[1][0] * point[0] + mat[1][1] * point[1]
    point[0] = x
    point[1] = y


def norm(point):
    """Normalise"""
    return sqrt(point[0] * point[0] + point[1] * point[1])



#!/usr/bin/env python
"""List all colors used in an svg document"""

from collections import defaultdict
from typing import Dict
import inkex


class ListColours(inkex.ColorExtension):
    """Make the colours darker"""

    _counts: Dict = defaultdict(int)

    def effect(self):
        super().effect()
        items = sorted(self._counts.items(), key=lambda v: -v[1])
        for color, count in items:
            self.msg("{count}: {color}".format(color=color, count=count))

    def modify_color(self, name, color):
        self._counts[color] += 1
        return color


if __name__ == "__main__":
    ListColours().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (c) 2009 - Jos Hirth, kaioa.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Export a gimp pallet file (.gpl)
"""

import inkex
from inkex import ShapeElement, ColorIdError, ColorError


class ExportGimpPalette(inkex.OutputExtension):
    """Export all colors in a document to a gimp pallet"""

    select_all = (ShapeElement,)
    names = {}

    def save(self, stream):
        name = self.svg.name.replace(".svg", "")
        stream.write("GIMP Palette\nName: {}\n#\n".format(name).encode("utf-8"))

        for key, value in sorted(list(set(self.get_colors()))):
            stream.write("{} {}\n".format(key, value).encode("utf-8"))

    def get_colors(self):
        """Get all the colors from the selected elements"""
        for elem in self.svg.selection.filter(ShapeElement):
            for color in self.process_element(elem):
                if str(color).upper() == "NONE":
                    continue
                yield (
                    "{:3d} {:3d} {:3d}".format(*color.to_rgb()),
                    self.names.get(color) or str(color).upper(),
                )

    def process_element(self, elem):
        """Recursively process elements for colors"""
        style = elem.specified_style()
        for col in inkex.Style.color_props:
            try:
                col = inkex.Color(style.get(col))
                if elem.getparent().get("inkscape:swatch") == "solid":
                    self.names[col] = elem.getparent().get_id()
                yield col
            except ColorIdError:
                gradient = self.svg.getElementById(style.get(col))
                for stop in gradient.stops:
                    for item in self.process_element(stop):
                        yield item
            except ColorError:
                pass  # Bad color

        if (
            elem.href is not None
        ):  # Capture colors of symbols or clones pointing to defs
            for color in self.process_element(elem.href):
                yield color


if __name__ == "__main__":
    ExportGimpPalette().run()



#!/usr/bin/env python
# coding=utf-8
#
# Authors:
#   Nicolas Dufour - Association Inkscape-fr
#   Aurelio A. Heckert <aurium(a)gmail.com>
#
# Copyright (C) 2008 Authors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension allows you to draw crop, registration and other
printing marks in Inkscape.
"""

import math
import inkex
from inkex import Circle, Rectangle, TextElement


class PrintingMarks(inkex.EffectExtension):
    # Default parameters
    stroke_width = 0.25

    def add_arguments(self, pars):
        pars.add_argument("--where", help="Apply crop marks to...", default="canvas")
        pars.add_argument(
            "--crop_marks", type=inkex.Boolean, default=True, help="Draw crop Marks"
        )
        pars.add_argument("--bleed_marks", type=inkex.Boolean, help="Draw Bleed Marks")
        pars.add_argument(
            "--registration_marks",
            type=inkex.Boolean,
            dest="reg_marks",
            default=True,
            help="Draw Registration Marks?",
        )
        pars.add_argument("--star_target", type=inkex.Boolean, help="Draw Star Target?")
        pars.add_argument(
            "--colour_bars", type=inkex.Boolean, help="Draw Colour Bars?", default=True
        )
        pars.add_argument(
            "--page_info", type=inkex.Boolean, help="Draw Page Information?"
        )
        pars.add_argument("--unit", default="px", help="Draw measurement")
        pars.add_argument("--crop_offset", type=float, default=5.0, help="Offset")
        pars.add_argument("--bleed_top", type=float, default=5.0, help="Bleed Top Size")
        pars.add_argument(
            "--bleed_bottom", type=float, default=5.0, help="Bleed Bottom Size"
        )
        pars.add_argument(
            "--bleed_left", type=float, default=5.0, help="Bleed Left Size"
        )
        pars.add_argument(
            "--bleed_right", type=float, default=5.0, help="Bleed Right Size"
        )
        pars.add_argument("--tab", help="The selected UI-tab when OK was pressed")

    def draw_crop_line(self, x1, y1, x2, y2, name, parent):
        style = {
            "stroke": "#000000",
            "stroke-width": str(self.stroke_width),
            "fill": "none",
        }
        line_attribs = {
            "style": str(inkex.Style(style)),
            "id": name,
            "d": "M " + str(x1) + "," + str(y1) + " L " + str(x2) + "," + str(y2),
        }
        parent.add(inkex.PathElement(**line_attribs))

    def draw_bleed_line(self, x1, y1, x2, y2, name, parent):
        style = {
            "stroke": "#000000",
            "stroke-width": str(self.stroke_width),
            "fill": "none",
            "stroke-miterlimit": "4",
            "stroke-dasharray": "4, 2, 1, 2",
            "stroke-dashoffset": "0",
        }
        line_attribs = {
            "style": str(inkex.Style(style)),
            "id": name,
            "d": "M " + str(x1) + "," + str(y1) + " L " + str(x2) + "," + str(y2),
        }
        parent.add(inkex.PathElement(**line_attribs))

    def draw_reg_circles(self, cx, cy, r, name, colours, parent):
        for i in range(len(colours)):
            style = {
                "stroke": colours[i],
                "stroke-width": str(r / len(colours)),
                "fill": "none",
            }
            circle_attribs = {
                "style": str(inkex.Style(style)),
                "inkscape:label": name,
                "cx": str(cx),
                "cy": str(cy),
                "r": str((r / len(colours)) * (i + 0.5)),
            }
            parent.add(Circle(**circle_attribs))

    def draw_reg_marks(self, cx, cy, rotate, name, parent):
        colours = ["#000000", "#00ffff", "#ff00ff", "#ffff00", "#000000"]
        g = parent.add(inkex.Group(id=name))
        for i in range(len(colours)):
            style = {"fill": colours[i], "fill-opacity": "1", "stroke": "none"}
            r = self.mark_size / 2
            step = r
            stroke = r / len(colours)
            regoffset = stroke * i
            regmark_attribs = {
                "style": str(inkex.Style(style)),
                "d": "m"
                + " "
                + str(-regoffset)
                + ","
                + str(r)
                + " "
                + str(-stroke)
                + ",0"
                + " "
                + str(step)
                + ","
                + str(-r)
                + " "
                + str(-step)
                + ","
                + str(-r)
                + " "
                + str(stroke)
                + ",0"
                + " "
                + str(step)
                + ","
                + str(r)
                + " "
                + str(-step)
                + ","
                + str(r)
                + " z",
                "transform": "translate("
                + str(cx)
                + ","
                + str(cy)
                + ") rotate("
                + str(rotate)
                + ")",
            }
            g.add(inkex.PathElement(**regmark_attribs))

    def draw_star_target(self, cx, cy, name, parent):
        r = self.mark_size / 2
        style = {
            "fill": "#000 device-cmyk(1,1,1,1)",
            "fill-opacity": "1",
            "stroke": "none",
        }
        d = " M 0,0"
        i = 0
        while i < (2 * math.pi):
            i += math.pi / 16
            d += (
                " L 0,0 "
                + " L "
                + str(math.sin(i) * r)
                + ","
                + str(math.cos(i) * r)
                + " L "
                + str(math.sin(i + 0.09) * r)
                + ","
                + str(math.cos(i + 0.09) * r)
            )
        elem = parent.add(inkex.PathElement())
        elem.label = name
        elem.transform.add_translate(cx, cy)
        elem.path = d
        elem.style = inkex.Style(style)

    def draw_coluor_bars(self, cx, cy, rotate, name, parent, bbox):
        group = parent.add(inkex.Group(id=name))
        group.transform = inkex.Transform(translate=(cx, cy)) @ inkex.Transform(
            rotate=rotate
        )
        loc = 0
        if bbox:
            loc = min(self.mark_size / 3, max(bbox.width, bbox.height) / 45)
        for bar in [
            {"c": "*", "stroke": "#000", "x": 0, "y": -(loc + 1)},
            {"c": "r", "stroke": "#0FF", "x": 0, "y": 0},
            {"c": "g", "stroke": "#F0F", "x": (loc * 11) + 1, "y": -(loc + 1)},
            {"c": "b", "stroke": "#FF0", "x": (loc * 11) + 1, "y": 0},
        ]:
            i = 0
            while i <= 1:
                color = inkex.Color("white")
                if bar["c"] == "r" or bar["c"] == "*":
                    color.red = 255 * i
                if bar["c"] == "g" or bar["c"] == "*":
                    color.green = 255 * i
                if bar["c"] == "b" or bar["c"] == "*":
                    color.blue = 255 * i
                r_att = {
                    "fill": str(color),
                    "stroke": bar["stroke"],
                    "stroke-width": loc / 8,
                    "x": str((loc * i * 10) + bar["x"]),
                    "y": str(bar["y"]),
                    "width": loc,
                    "height": loc,
                }
                rect = Rectangle()
                for att, value in r_att.items():
                    rect.set(att, value)
                group.add(rect)
                i += 0.1

    def effect(self):
        self.mark_size = self.svg.viewport_to_unit("1cm")
        self.min_mark_margin = self.svg.viewport_to_unit("3mm")

        if self.options.where == "selection":
            bbox = self.svg.selection.bounding_box()
            if bbox is None:
                raise inkex.AbortExtension(_("Selection is empty"))
        else:
            bbox = self.svg.get_page_bbox()

        # Get SVG document dimensions
        # self.width must be replaced by bbox.right. same to others.
        svg = self.document.getroot()

        # Convert parameters to user unit
        offset = self.svg.viewport_to_unit(
            str(self.options.crop_offset) + self.options.unit
        )
        bt = self.svg.viewport_to_unit(str(self.options.bleed_top) + self.options.unit)
        bb = self.svg.viewport_to_unit(
            str(self.options.bleed_bottom) + self.options.unit
        )
        bl = self.svg.viewport_to_unit(str(self.options.bleed_left) + self.options.unit)
        br = self.svg.viewport_to_unit(
            str(self.options.bleed_right) + self.options.unit
        )
        # Bleed margin
        if bt < offset:
            bmt = 0
        else:
            bmt = bt - offset
        if bb < offset:
            bmb = 0
        else:
            bmb = bb - offset
        if bl < offset:
            bml = 0
        else:
            bml = bl - offset
        if br < offset:
            bmr = 0
        else:
            bmr = br - offset

        # Define the new document limits
        offset_left = bbox.left - offset
        offset_right = bbox.right + offset
        offset_top = bbox.top - offset
        offset_bottom = bbox.bottom + offset

        # Get middle positions
        middle_vertical = bbox.top + (bbox.height / 2)
        middle_horizontal = bbox.left + (bbox.width / 2)

        # Test if printing-marks layer existis
        layer = self.svg.xpath(
            '//*[@id="printing-marks" and @inkscape:groupmode="layer"]'
        )
        if layer:
            svg.remove(layer[0])  # remove if it existis
        # Create a new layer
        layer = svg.add(inkex.Layer.new("Printing Marks"))
        layer.set("id", "printing-marks")
        layer.set("sodipodi:insensitive", "true")

        # Crop Mark
        if self.options.crop_marks:
            # Create a group for Crop Mark
            g_crops = layer.add(inkex.Group(id="CropMarks"))
            g_crops.label = "CropMarks"

            # Top left Mark
            self.draw_crop_line(
                bbox.left,
                offset_top,
                bbox.left,
                offset_top - self.mark_size,
                "cropTL1",
                g_crops,
            )
            self.draw_crop_line(
                offset_left,
                bbox.top,
                offset_left - self.mark_size,
                bbox.top,
                "cropTL2",
                g_crops,
            )

            # Top right Mark
            self.draw_crop_line(
                bbox.right,
                offset_top,
                bbox.right,
                offset_top - self.mark_size,
                "cropTR1",
                g_crops,
            )
            self.draw_crop_line(
                offset_right,
                bbox.top,
                offset_right + self.mark_size,
                bbox.top,
                "cropTR2",
                g_crops,
            )

            # Bottom left Mark
            self.draw_crop_line(
                bbox.left,
                offset_bottom,
                bbox.left,
                offset_bottom + self.mark_size,
                "cropBL1",
                g_crops,
            )
            self.draw_crop_line(
                offset_left,
                bbox.bottom,
                offset_left - self.mark_size,
                bbox.bottom,
                "cropBL2",
                g_crops,
            )

            # Bottom right Mark
            self.draw_crop_line(
                bbox.right,
                offset_bottom,
                bbox.right,
                offset_bottom + self.mark_size,
                "cropBR1",
                g_crops,
            )
            self.draw_crop_line(
                offset_right,
                bbox.bottom,
                offset_right + self.mark_size,
                bbox.bottom,
                "cropBR2",
                g_crops,
            )

        # Bleed Mark
        if self.options.bleed_marks:
            # Create a group for Bleed Mark
            g_bleed = layer.add(inkex.Group())
            g_bleed.label = "BleedMarks"
            g_bleed.set("id", "BleedMarks")

            # Top left Mark
            self.draw_bleed_line(
                bbox.left - bl,
                offset_top - bmt,
                bbox.left - bl,
                offset_top - bmt - self.mark_size,
                "bleedTL1",
                g_bleed,
            )
            self.draw_bleed_line(
                offset_left - bml,
                bbox.top - bt,
                offset_left - bml - self.mark_size,
                bbox.top - bt,
                "bleedTL2",
                g_bleed,
            )

            # Top right Mark
            self.draw_bleed_line(
                bbox.right + br,
                offset_top - bmt,
                bbox.right + br,
                offset_top - bmt - self.mark_size,
                "bleedTR1",
                g_bleed,
            )
            self.draw_bleed_line(
                offset_right + bmr,
                bbox.top - bt,
                offset_right + bmr + self.mark_size,
                bbox.top - bt,
                "bleedTR2",
                g_bleed,
            )

            # Bottom left Mark
            self.draw_bleed_line(
                bbox.left - bl,
                offset_bottom + bmb,
                bbox.left - bl,
                offset_bottom + bmb + self.mark_size,
                "bleedBL1",
                g_bleed,
            )
            self.draw_bleed_line(
                offset_left - bml,
                bbox.bottom + bb,
                offset_left - bml - self.mark_size,
                bbox.bottom + bb,
                "bleedBL2",
                g_bleed,
            )

            # Bottom right Mark
            self.draw_bleed_line(
                bbox.right + br,
                offset_bottom + bmb,
                bbox.right + br,
                offset_bottom + bmb + self.mark_size,
                "bleedBR1",
                g_bleed,
            )
            self.draw_bleed_line(
                offset_right + bmr,
                bbox.bottom + bb,
                offset_right + bmr + self.mark_size,
                bbox.bottom + bb,
                "bleedBR2",
                g_bleed,
            )

        # Registration Mark
        if self.options.reg_marks:
            # Create a group for Registration Mark
            g_center = layer.add(inkex.Group())
            g_center.label = "RegistrationMarks"
            g_center.set("id", "RegistrationMarks")
            # Left Mark
            cx = max(bml + offset, self.min_mark_margin)
            self.draw_reg_marks(
                bbox.left - cx - (self.mark_size / 2),
                middle_vertical - self.mark_size * 1.5,
                "0",
                "regMarkL",
                g_center,
            )

            # Right Mark
            cx = max(bmr + offset, self.min_mark_margin)
            self.draw_reg_marks(
                bbox.right + cx + (self.mark_size / 2),
                middle_vertical - self.mark_size * 1.5,
                "180",
                "regMarkR",
                g_center,
            )

            # Top Mark
            cy = max(bmt + offset, self.min_mark_margin)
            self.draw_reg_marks(
                middle_horizontal,
                bbox.top - cy - (self.mark_size / 2),
                "90",
                "regMarkT",
                g_center,
            )

            # Bottom Mark
            cy = max(bmb + offset, self.min_mark_margin)
            self.draw_reg_marks(
                middle_horizontal,
                bbox.bottom + cy + (self.mark_size / 2),
                "-90",
                "regMarkB",
                g_center,
            )

        # Star Target
        if self.options.star_target:
            # Create a group for Star Target
            g_center = layer.add(inkex.Group())
            g_center.label = "StarTarget"
            g_center.set("id", "StarTarget")

            if bbox.height < bbox.width:
                # Left Star
                cx = max(bml + offset, self.min_mark_margin)
                self.draw_star_target(
                    bbox.left - cx - (self.mark_size / 2),
                    middle_vertical,
                    "starTargetL",
                    g_center,
                )
                # Right Star
                cx = max(bmr + offset, self.min_mark_margin)
                self.draw_star_target(
                    bbox.right + cx + (self.mark_size / 2),
                    middle_vertical,
                    "starTargetR",
                    g_center,
                )
            else:
                # Top Star
                cy = max(bmt + offset, self.min_mark_margin)
                self.draw_star_target(
                    middle_horizontal - self.mark_size * 1.5,
                    bbox.top - cy - (self.mark_size / 2),
                    "starTargetT",
                    g_center,
                )
                # Bottom Star
                cy = max(bmb + offset, self.min_mark_margin)
                self.draw_star_target(
                    middle_horizontal - self.mark_size * 1.5,
                    bbox.bottom + cy + (self.mark_size / 2),
                    "starTargetB",
                    g_center,
                )

        # Colour Bars
        if self.options.colour_bars:
            # Create a group for Colour Bars
            g_center = layer.add(inkex.Group())
            g_center.label = "ColourBars"
            g_center.set("id", "PrintingColourBars")

            if bbox.height > bbox.width:
                # Left Bars
                cx = max(bml + offset, self.min_mark_margin)
                self.draw_coluor_bars(
                    bbox.left - cx - (self.mark_size / 2),
                    middle_vertical + self.mark_size,
                    90,
                    "PrintingColourBarsL",
                    g_center,
                    bbox,
                )
                # Right Bars
                cx = max(bmr + offset, self.min_mark_margin)
                self.draw_coluor_bars(
                    bbox.right + cx + (self.mark_size / 2),
                    middle_vertical + self.mark_size,
                    90,
                    "PrintingColourBarsR",
                    g_center,
                    bbox,
                )
            else:
                # Top Bars
                cy = max(bmt + offset, self.min_mark_margin)
                self.draw_coluor_bars(
                    middle_horizontal + self.mark_size,
                    bbox.top - cy - (self.mark_size / 2),
                    0,
                    "PrintingColourBarsT",
                    g_center,
                    bbox,
                )
                # Bottom Bars
                cy = max(bmb + offset, self.min_mark_margin)
                self.draw_coluor_bars(
                    middle_horizontal + self.mark_size,
                    bbox.bottom + cy + (self.mark_size / 2),
                    0,
                    "PrintingColourBarsB",
                    g_center,
                    bbox,
                )

        # Page Information
        if self.options.page_info:
            # Create a group for Page Information
            g_pag_info = layer.add(inkex.Group())
            g_pag_info.label = "PageInformation"
            g_pag_info.set("id", "PageInformation")
            y_margin = max(bmb + offset, self.min_mark_margin)
            font_size = self.svg.viewport_to_unit("9pt")
            txt_attribs = {
                "style": f"font-size:{font_size}px;font-style:normal;font-weight:normal;fill:#000000;font-family:Bitstream Vera Sans,sans-serif;text-anchor:middle;text-align:center",
                "x": str(middle_horizontal),
                "y": str(bbox.bottom + y_margin + self.mark_size + 20),
            }
            txt = g_pag_info.add(TextElement(**txt_attribs))
            txt.text = (
                "Page size: "
                + str(
                    round(self.svg.unit_to_viewport(bbox.width, self.options.unit), 2)
                )
                + "x"
                + str(
                    round(self.svg.unit_to_viewport(bbox.height, self.options.unit), 2)
                )
                + " "
                + self.options.unit
            )


if __name__ == "__main__":
    PrintingMarks().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2013 <Public Domain>
#               2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import inkex
from inkex.ports import Serial
from inkex.localization import inkex_gettext as _

import hpgl_encoder


class Plot(inkex.EffectExtension):
    """Generate a plot in HPGL output"""

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument(
            "--parallelPort", default="/dev/usb/lp2", help="Parallel port"
        )
        pars.add_argument("--serialPort", default="COM1", help="Serial port")
        pars.add_argument("--serialBaudRate", default="9600", help="Serial Baud rate")
        pars.add_argument("--serialByteSize", default="eight", help="Serial byte size")
        pars.add_argument("--serialStopBits", default="one", help="Serial stop bits")
        pars.add_argument("--serialParity", default="none", help="Serial parity")
        pars.add_argument("--serialFlowControl", default="xonxoff", help="Flow control")
        pars.add_argument(
            "--resolutionX", type=float, default=1016.0, help="Resolution X (dpi)"
        )
        pars.add_argument(
            "--resolutionY", type=float, default=1016.0, help="Resolution Y (dpi)"
        )
        pars.add_argument("--pen", type=int, default=1, help="Pen number")
        pars.add_argument("--force", type=int, default=0, help="Pen force (g)")
        pars.add_argument("--speed", type=int, default=0, help="Pen speed (cm/s)")
        pars.add_argument("--orientation", default="0", help="Rotation (Clockwise)")
        pars.add_argument(
            "--mirrorX", type=inkex.Boolean, default=False, help="Mirror X axis"
        )
        pars.add_argument(
            "--mirrorY", type=inkex.Boolean, default=False, help="Mirror Y axis"
        )
        pars.add_argument(
            "--center", type=inkex.Boolean, default=False, help="Center zero point"
        )
        pars.add_argument("--overcut", type=float, default=1.0, help="Overcut (mm)")
        pars.add_argument(
            "--precut", type=inkex.Boolean, default=True, help="Use precut"
        )
        pars.add_argument("--flat", type=float, default=1.2, help="Curve flatness")
        pars.add_argument(
            "--autoAlign", type=inkex.Boolean, default=True, help="Auto align"
        )
        pars.add_argument(
            "--toolOffset",
            type=float,
            default=0.25,
            help="Tool (Knife) offset correction (mm)",
        )
        pars.add_argument(
            "--portType",
            type=self.arg_method("to"),
            default=self.to_serial,
            dest="to_port",
            help="Port type",
        )
        pars.add_argument(
            "--commandLanguage",
            type=self.arg_method("convert"),
            default=self.convert_hpgl,
            dest="to_language",
            help="Command Language Filter",
        )

    def effect(self):
        # get hpgl data
        encoder = hpgl_encoder.hpglEncoder(self)
        try:
            self.options.to_port(self.options.to_language(encoder.getHpgl()))
        except hpgl_encoder.NoPathError:
            raise inkex.AbortExtension(
                _("No paths where found. Please convert objects to paths.")
            )

    def convert_hpgl(self, hpgl):
        """Convert raw HPGL to HPGL"""
        init = "IN"
        # if self.options.force > 0:
        #    init += ';FS%d' % self.options.force
        # if self.options.speed > 0:
        #    init += ';VS%d' % self.options.speed
        return init + hpgl + ";PU0,0;SP0;IN; "

    def convert_dmpl(self, hpgl):
        """Convert HPGL to DMPL"""
        # ;: = Initialise plotter
        # H = Home position
        # A = Absolute pen positioning
        # Ln = Line type
        # Pn = Pen select
        # Vn = velocity
        # ECn = Coordinate addressing, 1: 0.001 inch, 5: 0.005 inch, M: 0.1 mm
        # D = Pen down
        # U = Pen up
        # Z = Reset plotter
        # n,n, = Coordinate pair
        hpgl = hpgl.replace(";", ",")
        hpgl = hpgl.replace("SP", "P")
        hpgl = hpgl.replace("PU", "U")
        hpgl = hpgl.replace("PD", "D")
        init = ";:HAL0"
        if self.options.speed > 0:
            init += "V%d" % self.options.speed
        init += "EC1"
        return init + hpgl[1:] + ",P0,U0,0,Z "

    def convert_knk(self, hpgl):
        """Convert HPGL to KNK Plotter Language"""
        init = "ZG"
        # if self.options.force > 0:
        #    init += ';FS%d' % self.options.force
        # if self.options.speed > 0:
        #    init += ';VS%d' % self.options.speed
        return init + hpgl + ";SP0;PU0,0;@ "

    def to_parallel(self, hpgl):
        """Output to hgpl to a parallel port"""
        port = open(self.options.parallelPort, "wb")
        port.write(hpgl.encode("utf8"))
        port.close()

    def to_serial(self, hpgl):
        """Output to hgpl to a serial port"""
        with Serial(
            self.options.serialPort,
            baud=self.options.serialBaudRate,
            stop=self.options.serialStopBits,
            size=self.options.serialByteSize,
            flow=self.options.serialFlowControl,
            parity=self.options.serialParity,
        ) as comx:
            comx.write(hpgl.encode("utf8"))


if __name__ == "__main__":
    Plot().run()



#!/usr/bin/env python
"""Convert to grey"""

import inkex


class Grayscale(inkex.ColorExtension):
    """Make all colours grayscale"""

    def modify_color(self, name, color):
        # ITU-R Recommendation BT.709 (NTSC and PAL)
        # l = 0.2125 * r + 0.7154 * g + 0.0721 * b
        lum = 0.299 * color.red + 0.587 * color.green + 0.114 * color.blue
        return inkex.Color((int(round(lum)), int(round(lum)), int(round(lum))))


if __name__ == "__main__":
    Grayscale().run()



#! /usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2009 Aurelio A. Heckert <aurium (a) gmail dot com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

__version__ = "0.2"

import inkex


class FoldableBox(inkex.EffectExtension):
    """Foldable Box generation."""

    def add_arguments(self, pars):
        pars.add_argument("--width", type=float, default=10.0, help="The Box Width")
        pars.add_argument("--height", type=float, default=15.0, help="The Box Height")
        pars.add_argument(
            "--depth", type=float, default=3.0, help="The Box Depth (z dimention)"
        )
        pars.add_argument("--unit", default="px", help="The unit of the box dimensions")
        pars.add_argument(
            "--proportion", type=float, default=0.6, help="Inner tab proportion"
        )
        pars.add_argument(
            "--guide", type=inkex.Boolean, default=False, help="Add guide lines"
        )

    def guide(self, value, orient):
        """Create a guideline conditionally"""
        if self.options.guide:
            self.svg.namedview.new_guide(value, orient)

    def effect(self):
        doc_w = self.svg.unittouu(self.document.getroot().get("width"))
        doc_h = self.svg.unittouu(self.document.getroot().get("height"))

        box_w = self.svg.unittouu(str(self.options.width) + self.options.unit)
        box_h = self.svg.unittouu(str(self.options.height) + self.options.unit)
        box_d = self.svg.unittouu(str(self.options.depth) + self.options.unit)
        tab_h = box_d * self.options.proportion

        box_id = self.svg.get_unique_id("box")
        group = self.svg.get_current_layer().add(inkex.Group(id=box_id))

        line_style = {
            "stroke": "#000000",
            "fill": "none",
            "stroke-width": str(self.svg.unittouu("1px")),
        }

        self.guide(doc_h, True)

        # Inner Close Tab
        line = group.add(inkex.PathElement(id=box_id + "-inner-close-tab"))
        line.path = [
            ["M", [box_w - (tab_h * 0.7), 0]],
            ["C", [box_w - (tab_h * 0.25), 0, box_w, tab_h * 0.3, box_w, tab_h * 0.9]],
            ["L", [box_w, tab_h]],
            ["L", [0, tab_h]],
            ["L", [0, tab_h * 0.9]],
            ["C", [0, tab_h * 0.3, tab_h * 0.25, 0, tab_h * 0.7, 0]],
            ["Z", []],
        ]
        line.style = line_style

        lower_pos = box_d + tab_h
        left_pos = 0

        self.guide(doc_h - tab_h, True)

        # Upper Close Tab
        line = group.add(inkex.PathElement(id=box_id + "-upper-close-tab"))
        line.path = [
            ["M", [left_pos, tab_h]],
            ["L", [left_pos + box_w, tab_h]],
            ["L", [left_pos + box_w, lower_pos]],
            ["L", [left_pos + 0, lower_pos]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_w

        # Upper Right Tab
        side_tab_h = lower_pos - (box_w / 2)
        if side_tab_h < tab_h:
            side_tab_h = tab_h

        line = group.add(inkex.PathElement(id=box_id + "-upper-right-tab"))
        line.path = [
            ["M", [left_pos, side_tab_h]],
            ["L", [left_pos + (box_d * 0.8), side_tab_h]],
            ["L", [left_pos + box_d, ((lower_pos * 3) - side_tab_h) / 3]],
            ["L", [left_pos + box_d, lower_pos]],
            ["L", [left_pos + 0, lower_pos]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_w + box_d

        # Upper Left Tab
        line = group.add(inkex.PathElement(id=box_id + "-upper-left-tab"))
        line.path = [
            ["M", [left_pos + box_d, side_tab_h]],
            ["L", [left_pos + (box_d * 0.2), side_tab_h]],
            ["L", [left_pos, ((lower_pos * 3) - side_tab_h) / 3]],
            ["L", [left_pos, lower_pos]],
            ["L", [left_pos + box_d, lower_pos]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos = 0

        self.guide(doc_h - tab_h - box_d, True)

        # Right Tab
        line = group.add(inkex.PathElement(id=box_id + "-left-tab"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos - (box_d / 2), lower_pos + (box_d / 4)]],
            ["L", [left_pos - (box_d / 2), lower_pos + box_h - (box_d / 4)]],
            ["L", [left_pos, lower_pos + box_h]],
            ["Z", []],
        ]
        line.style = line_style

        # Front
        line = group.add(inkex.PathElement(id=box_id + "-front"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos + box_w, lower_pos]],
            ["L", [left_pos + box_w, lower_pos + box_h]],
            ["L", [left_pos, lower_pos + box_h]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_w

        # Right
        line = group.add(inkex.PathElement(id=box_id + "-right"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos + box_d, lower_pos]],
            ["L", [left_pos + box_d, lower_pos + box_h]],
            ["L", [left_pos, lower_pos + box_h]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_d

        # Back
        line = group.add(inkex.PathElement(id=box_id + "-back"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos + box_w, lower_pos]],
            ["L", [left_pos + box_w, lower_pos + box_h]],
            ["L", [left_pos, lower_pos + box_h]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_w

        # Left
        line = group.add(inkex.PathElement(id=box_id + "-line"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos + box_d, lower_pos]],
            ["L", [left_pos + box_d, lower_pos + box_h]],
            ["L", [left_pos, lower_pos + box_h]],
            ["Z", []],
        ]
        line.style = line_style

        lower_pos += box_h
        left_pos = 0
        b_tab = lower_pos + box_d
        if b_tab > box_w / 2.5:
            b_tab = box_w / 2.5

        # Bottom Front Tab
        line = group.add(inkex.PathElement(id=box_id + "-bottom-front-tab"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos, lower_pos + (box_d / 2)]],
            ["L", [left_pos + box_w, lower_pos + (box_d / 2)]],
            ["L", [left_pos + box_w, lower_pos]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_w

        # Bottom Right Tab
        line = group.add(inkex.PathElement(id=box_id + "-bottom-right-tab"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos, lower_pos + b_tab]],
            ["L", [left_pos + box_d, lower_pos + b_tab]],
            ["L", [left_pos + box_d, lower_pos]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_d

        # Bottom Back Tab
        line = group.add(inkex.PathElement(id=box_id + "-bottom-back-tab"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos, lower_pos + (box_d / 2)]],
            ["L", [left_pos + box_w, lower_pos + (box_d / 2)]],
            ["L", [left_pos + box_w, lower_pos]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_w

        # Bottom Left Tab
        line = group.add(inkex.PathElement(id=box_id + "-bottom-left-tab"))
        line.path = [
            ["M", [left_pos, lower_pos]],
            ["L", [left_pos, lower_pos + b_tab]],
            ["L", [left_pos + box_d, lower_pos + b_tab]],
            ["L", [left_pos + box_d, lower_pos]],
            ["Z", []],
        ]
        line.style = line_style

        left_pos += box_d
        lower_pos += b_tab

        group.transform = inkex.Transform(
            translate=((doc_w - left_pos) / 2, (doc_h - lower_pos) / 2)
        )


if __name__ == "__main__":  # pragma: no cover
    FoldableBox().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) Aaron Spike <aaron@ekips.org>
#               Aurélio A. Heckert <aurium(a)gmail.com>
#               Bulia Byak <buliabyak@users.sf.net>
#               Nicolas Dufour, nicoduf@yahoo.fr
#               Peter J. R. Moulder <pjrm@users.sourceforge.net>
#               Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Convert to and from various units and find the closest matching unit.
"""

import re

# a dictionary of unit to user unit conversion factors
CONVERSIONS = {
    "in": 96.0,
    "pt": 1.3333333333333333,
    "px": 1.0,
    "mm": 3.779527559055118,
    "cm": 37.79527559055118,
    "m": 3779.527559055118,
    "km": 3779527.559055118,
    "Q": 0.94488188976378,
    "pc": 16.0,
    "yd": 3456.0,
    "ft": 1152.0,
    "": 1.0,  # Default px
}

# allowed unit types, including percentages, relative units, and others
# that are not suitable for direct conversion to a length.
# Note that this is _not_ an exhaustive list of allowed unit types.
UNITS = [
    "in",
    "pt",
    "px",
    "mm",
    "cm",
    "m",
    "km",
    "Q",
    "pc",
    "yd",
    "ft",
    "",
    "%",
    "em",
    "ex",
    "ch",
    "rem",
    "vw",
    "vh",
    "vmin",
    "vmax",
    "deg",
    "grad",
    "rad",
    "turn",
    "s",
    "ms",
    "Hz",
    "kHz",
    "dpi",
    "dpcm",
    "dppx",
]

UNIT_MATCH = re.compile(rf"({'|'.join(UNITS)})")
NUMBER_MATCH = re.compile(r"(([-+]?[0-9]+(\.[0-9]*)?|[-+]?\.[0-9]+)([eE][-+]?[0-9]+)?)")
BOTH_MATCH = re.compile(rf"^\s*{NUMBER_MATCH.pattern}\s*{UNIT_MATCH.pattern}\s*$")


def parse_unit(value, default_unit="px", default_value=None):
    """
    Takes a value such as 55.32px and returns (55.32, 'px')
    Returns default (None) if no match can be found
    """
    ret = BOTH_MATCH.match(str(value))
    if ret:
        return float(ret.groups()[0]), ret.groups()[-1] or default_unit
    return (default_value, default_unit) if default_value is not None else None


def are_near_relative(point_a, point_b, eps=0.01):
    """Return true if the points are near to eps"""
    return (point_a - point_b <= point_a * eps) and (
        point_a - point_b >= -point_a * eps
    )


def discover_unit(value, viewbox, default="px"):
    """Attempt to detect the unit being used based on the viewbox"""
    # Default 100px when width can't be parsed
    (value, unit) = parse_unit(value, default_value=100.0)
    if unit not in CONVERSIONS:
        return default
    this_factor = CONVERSIONS[unit] * value / viewbox

    # try to find the svgunitfactor in the list of units known. If we don't find
    # something, ...
    for unit, unit_factor in CONVERSIONS.items():
        if unit != "":
            # allow 1% error in factor
            if are_near_relative(this_factor, unit_factor, eps=0.01):
                return unit
    return default


def convert_unit(value, to_unit, default="px"):
    """Returns userunits given a string representation of units in another system

    Args:
        value: <length> string
        to_unit: unit to convert to
        default: if ``value`` contains no unit, what unit should be assumed.

            .. versionadded:: 1.1
    """
    value, from_unit = parse_unit(value, default_unit=default, default_value=0.0)
    if from_unit in CONVERSIONS and to_unit in CONVERSIONS:
        return (
            value * CONVERSIONS[from_unit] / CONVERSIONS.get(to_unit, CONVERSIONS["px"])
        )
    return 0.0


def render_unit(value, unit):
    """Checks and then renders a number with its unit"""
    try:
        if isinstance(value, str):
            (value, unit) = parse_unit(value, default_unit=unit)
        return f"{value:.6g}{ unit:s}"
    except TypeError:
        return ""



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (c) 2009 Bryan Hoyt (MIT License)
#               2011 Nicolas Dufour <nicoduf@yahoo.fr>
#               2013 Johan B. C. Engelen <j.b.c.engelen@alumnus.utwente.nl>
#               2019 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
TODO: mark elements that have previously been snapped, along with the settings
    used, so that the same settings can be used for that element next time when
    it's selected as part of a group (and add an option to the extension dialog
    "Use previous/default settings" which is selected by default)

TODO: make elem_offset return [x_offset, y_offset] so we can handle non-symetric scaling
      => will probably need to take into account non-symetric scaling on stroke-widths,
         too (horizontal vs vertical strokes)

TODO: Transforming points isn't quite perfect, to say the least. In particular,
    when translating a point on a bezier curve, we translate the handles by the same amount.
    BUT, some handles that are attached to a particular point are conceptually
    handles of the prev/next node.
    Best way to fix it would be to keep a list of the fractional_offsets[] of
    each point, without transforming anything. Then go through each point and
    transform the appropriate handle according to the relevant fraction_offset
    in the list.

    i.e. calculate first, then modify.

    In fact, that might be a simpler algorithm anyway -- it avoids having
    to keep track of all the first_xy/next_xy guff.

Note: This doesn't work very well on paths which have both straight segments
      and curved segments.
      The biggest three problems are:
        a) we don't take handles into account (segments where the nodes are
           aligned are always treated as straight segments, even where the
           handles make it curve)
        b) when we snap a straight segment right before/after a curve, it
           doesn't make any attempt to keep the transition from the straight
           segment to the curve smooth.
        c) no attempt is made to keep equal widths equal. (or nearly-equal
           widths nearly-equal). For example, font strokes.

Note: Paths that have curves & arcs on some sides of the bounding box won't
    be snapped correctly on that side of the bounding box, and nor will they
    be translated/resized correctly before the path is modified. Doesn't affect
    most applications of this extension, but it highlights the fact that we
    take a geometrically simplistic approach to inspecting & modifying the path.
"""

from __future__ import division
from __future__ import print_function

import sys

import inkex
from inkex import PathElement, Group, Image, Rectangle, ShapeElement, Transform
from inkex.localization import inkex_gettext as _

Precision = 5  # number of digits of precision for comparing float numbers


class TransformError(Exception):
    pass


def transform_point(transform, pt, inverse=False):
    """apply_to_point with inbuilt inverse"""
    if inverse:
        transform = -transform
    return transform.apply_to_point(pt)


def transform_dimensions(transform, width=None, height=None, inverse=False):
    """Dimensions don't get translated. I'm not sure how much diff rotate/skew
    makes in this context, but we currently ignore anything besides scale.
    """
    if inverse:
        transform = -transform

    if width is not None:
        width *= transform.a
    if height is not None:
        height *= transform.d

    if width is not None and height is not None:
        return width, height
    if width is not None:
        return width
    if height is not None:
        return height


class PixelSnap(inkex.EffectExtension):
    """Snap objects to pixels"""

    def add_arguments(self, pars):
        """Add inx options"""
        pars.add_argument(
            "-a",
            "--snap_ancestors",
            type=inkex.Boolean,
            default=True,
            help="Snap unselected ancestors' translations "
            "(groups, layers, document height) first",
        )
        pars.add_argument(
            "-t",
            "--ancestor_offset",
            type=inkex.Boolean,
            default=True,
            help="Calculate offset relative to unselected ancestors' "
            "transforms (includes document height offset)",
        )
        pars.add_argument(
            "-g",
            "--max_gradient",
            type=float,
            default=0.5,
            help="Maximum slope to consider straight (%)",
        )
        pars.add_argument(
            "-s",
            "--snap_to",
            default="tl",
            choices=["tl", "bl"],
            help="Origin of the coordinate system",
        )

    def vertical(self, pt1, pt2):
        hlen = abs(pt1[0] - pt2[0])
        vlen = abs(pt1[1] - pt2[1])
        if vlen == 0 and hlen == 0:
            return True
        elif vlen == 0:
            return False
        return (hlen / vlen) < self.options.max_gradient / 100

    def horizontal(self, pt1, pt2):
        hlen = round(abs(pt1[0] - pt2[0]), Precision)
        vlen = round(abs(pt1[1] - pt2[1]), Precision)
        if hlen == 0 and vlen == 0:
            return True
        elif hlen == 0:
            return False
        return (vlen / hlen) < self.options.max_gradient / 100

    def stroke_width_offset(self, elem, parent_transform=None):
        """Returns the amount the bounding-box is offset due to the stroke-width.
        Transform is taken into account.
        """
        stroke_width = self.stroke_width(elem)
        if stroke_width == 0:
            return 0  # if there's no stroke, no need to worry about the transform

        transform = elem.transform @ Transform(parent_transform)

        if abs(abs(transform.a) - abs(transform.d)) > (10**-Precision):
            raise TransformError(
                _("Selection contains non-symetric scaling")
            )  # *** wouldn't be hard to get around this by calculating vertical_offset & horizontal_offset separately, maybe 1 functions, or maybe returning a tuple

        stroke_width = transform_dimensions(transform, width=stroke_width)

        return stroke_width / 2

    def stroke_width(self, elem, setval=None):
        """Get/set stroke-width in pixels, untransformed"""
        style = elem.style
        stroke = style("stroke")

        stroke_width = 0
        if stroke and setval is None:
            stroke_width = self.svg.to_dimensionless(style("stroke-width").strip())

        if setval:
            style["stroke-width"] = setval
        else:
            return stroke_width

    def transform_path_node(self, transform, path, i):
        """Modifies a segment so that every point is transformed, including handles"""
        segtype = path[i][0].lower()

        if segtype == "z":
            return
        elif segtype == "h":
            path[i][1][0] = transform_point(transform, [path[i][1][0], 0])[0]
        elif segtype == "v":
            path[i][1][0] = transform_point(transform, [0, path[i][1][0]])[1]
        else:
            first_coordinate = 0
            if segtype == "a":
                first_coordinate = 5  # for elliptical arcs, skip the radius x/y, rotation, large-arc, and sweep
            for j in range(first_coordinate, len(path[i][1]), 2):
                x, y = path[i][1][j], path[i][1][j + 1]
                x, y = transform_point(transform, (x, y))
                path[i][1][j] = x
                path[i][1][j + 1] = y

    def pathxy(self, path, i, setval=None):
        """Return the endpoint of the given path segment.
        Inspects the segment type to know which elements are the endpoints.
        """
        segtype = path[i][0].lower()
        x = y = 0

        if segtype == "z":
            i = 0

        if segtype == "h":
            if setval:
                path[i][1][0] = setval[0]
            else:
                x = path[i][1][0]

        elif segtype == "v":
            if setval:
                path[i][1][0] = setval[1]
            else:
                y = path[i][1][0]
        else:
            if setval and segtype != "z":
                path[i][1][-2] = setval[0]
                path[i][1][-1] = setval[1]
            else:
                x = path[i][1][-2]
                y = path[i][1][-1]

        if setval is None:
            return [x, y]

    def snap_path_scale(self, elem, parent_transform=None):

        path = elem.original_path.to_arrays()
        transform = elem.transform @ Transform(parent_transform)
        bbox = elem.bounding_box()

        # In case somebody tries to snap a 0-high element,
        # or a curve/arc with all nodes in a line, and of course
        # because we should always check for divide-by-zero!
        if not bbox.width or not bbox.height:
            return

        width, height = bbox.width, bbox.height
        min_xy, max_xy = bbox.minimum, bbox.maximum
        rescale = round(width) / width, round(height) / height

        min_xy = transform_point(transform, min_xy, inverse=True)
        max_xy = transform_point(transform, max_xy, inverse=True)

        for i in range(len(path)):
            translate = Transform(translate=min_xy)
            self.transform_path_node(-translate, path, i)  # center transform
            self.transform_path_node(Transform(scale=rescale), path, i)
            self.transform_path_node(translate, path, i)  # uncenter transform

        elem.original_path = path

    def snap_path_pos(self, elem, parent_transform=None):
        path = elem.original_path.to_arrays()
        transform = elem.transform @ Transform(parent_transform)
        bbox = elem.bounding_box()
        min_xy, max_xy = bbox.minimum, bbox.maximum

        fractional_offset = (
            min_xy[0] - round(min_xy[0]),
            min_xy[1] - round(min_xy[1]) - self.document_offset,
        )
        fractional_offset = transform_dimensions(
            transform, fractional_offset[0], fractional_offset[1], inverse=True
        )

        for i in range(len(path)):
            self.transform_path_node(-Transform(translate=fractional_offset), path, i)

        path = str(inkex.Path(path))
        if elem.get("inkscape:original-d"):
            elem.set("inkscape:original-d", path)
        else:
            elem.set("d", path)

    def snap_transform(self, elem):
        # Only snaps the x/y translation of the transform, nothing else.
        # Doesn't take any parent_transform into account -- assumes
        # that the parent's transform has already been snapped.
        transform = elem.transform
        # if we've got any skew/rotation, get outta here
        if transform.c or transform.b:
            raise TransformError(
                _("TR: Selection contains transformations with skew/rotation")
            )

        trm = list(transform.to_hexad())
        trm[4] = round(transform.e)
        trm[5] = round(transform.f)
        elem.transform @= Transform(trm)

    def snap_stroke(self, elem, parent_transform=None):
        transform = elem.transform @ Transform(parent_transform)

        stroke_width = self.stroke_width(elem)
        if stroke_width == 0:
            return  # no point raising a TransformError if there's no stroke to snap

        if abs(abs(transform.a) - abs(transform.d)) > (10**-Precision):
            raise TransformError(
                _("Selection contains non-symetric scaling, can't snap stroke width")
            )

        if stroke_width:
            stroke_width = transform_dimensions(transform, width=stroke_width)
            stroke_width = round(stroke_width)
            stroke_width = transform_dimensions(
                transform, width=stroke_width, inverse=True
            )
            self.stroke_width(elem, stroke_width)

    def snap_path(self, elem, parent_transform=None):
        path = elem.original_path.to_arrays()

        transform = elem.transform @ Transform(parent_transform)

        if transform.c or transform.b:  # if we've got any skew/rotation, get outta here
            raise TransformError(
                _("Path: Selection contains transformations with skew/rotation")
            )

        offset = self.stroke_width_offset(elem, parent_transform) % 1

        prev_xy = self.pathxy(path, -1)
        first_xy = self.pathxy(path, 0)
        for i in range(len(path)):
            segtype = path[i][0].lower()
            xy = self.pathxy(path, i)
            if segtype == "z":
                xy = first_xy
            if (i == len(path) - 1) or (
                (i == len(path) - 2) and path[-1][0].lower() == "z"
            ):
                next_xy = first_xy
            else:
                next_xy = self.pathxy(path, i + 1)

            if not (xy and prev_xy and next_xy):
                prev_xy = xy
                continue

            xy_untransformed = tuple(xy)
            xy = list(transform_point(transform, xy))
            prev_xy = transform_point(transform, prev_xy)
            next_xy = transform_point(transform, next_xy)

            on_vertical = on_horizontal = False

            if self.horizontal(xy, prev_xy):
                # on 2-point paths, first.next==first.prev==last and last.next==last.prev==first
                if len(path) > 2 or i == 0:
                    # make the almost-equal values equal, so they round in the same direction
                    xy[1] = prev_xy[1]
                on_horizontal = True
            if self.horizontal(xy, next_xy):
                on_horizontal = True

            if self.vertical(xy, prev_xy):  # as above
                if len(path) > 2 or i == 0:
                    xy[0] = prev_xy[0]
                on_vertical = True
            if self.vertical(xy, next_xy):
                on_vertical = True

            prev_xy = tuple(xy_untransformed)

            fractional_offset = [0, 0]
            if on_vertical:
                fractional_offset[0] = xy[0] - (round(xy[0] - offset) + offset)
            if on_horizontal:
                fractional_offset[1] = (
                    xy[1] - (round(xy[1] - offset) + offset) - self.document_offset
                )

            fractional_offset = transform_dimensions(
                transform, fractional_offset[0], fractional_offset[1], inverse=True
            )
            self.transform_path_node(-Transform(translate=fractional_offset), path, i)

        elem.original_path = path

    def snap_rect(self, elem, parent_transform=None):
        transform = elem.transform @ Transform(parent_transform)

        if transform.c or transform.b:  # if we've got any skew/rotation, get outta here
            raise TransformError(
                _("Rect: Selection contains transformations with skew/rotation")
            )

        offset = self.stroke_width_offset(elem, parent_transform) % 1

        width = self.svg.to_dimensionless(elem.attrib["width"])
        height = self.svg.to_dimensionless(elem.attrib["height"])
        x = self.svg.to_dimensionless(elem.attrib["x"])
        y = self.svg.to_dimensionless(elem.attrib["y"])

        width, height = transform_dimensions(transform, width, height)
        x, y = transform_point(transform, [x, y])

        # Snap to the nearest pixel
        height = round(height)
        width = round(width)
        x = (
            round(x - offset) + offset
        )  # If there's a stroke of non-even width, it's shifted by half a pixel
        y = round(y - offset) + offset

        width, height = transform_dimensions(transform, width, height, inverse=True)
        x, y = transform_point(transform, [x, y], inverse=True)
        y += self.document_offset / transform.d

        # Position the elem at the newly calculate values
        elem.attrib["width"] = str(width)
        elem.attrib["height"] = str(height)
        elem.attrib["x"] = str(x)
        elem.attrib["y"] = str(y)

    def snap_image(self, elem, parent_transform=None):
        self.snap_rect(elem, parent_transform)

    def pixel_snap(self, elem, parent_transform=None):
        if not isinstance(elem, (Group, Image, Rectangle, PathElement)):
            return

        if isinstance(elem, Group):
            self.snap_transform(elem)
            transform = elem.transform * Transform(parent_transform)
            for child in elem:
                try:
                    self.pixel_snap(child, transform)
                except TransformError as err:
                    raise inkex.AbortExtension(str(err))
            return

        # If we've been given a parent_transform, we can assume that the
        # parents have already been snapped, or don't need to be
        if self.options.snap_ancestors and parent_transform is None:
            # Loop through ancestors from outermost to innermost, excluding this element.
            for child in elem.ancestors():
                self.snap_transform(child)

        # If we haven't been given a parent_transform, then we need to calculate it
        if self.options.ancestor_offset and parent_transform is None:
            if isinstance(elem.getparent(), ShapeElement):
                parent_transform = elem.getparent().composed_transform()

        self.snap_transform(elem)
        try:
            self.snap_stroke(elem, parent_transform)
        except TransformError as err:
            raise inkex.AbortExtension(str(err))

        if isinstance(elem, PathElement):
            self.snap_path_scale(elem, parent_transform)
            self.snap_path_pos(elem, parent_transform)
            self.snap_path(
                elem, parent_transform
            )  # would be quite useful to make this an option, as scale/pos alone doesn't mess with the path itself, and works well for sans-serif text
        elif isinstance(elem, Rectangle):
            self.snap_rect(elem, parent_transform)
        elif isinstance(elem, Image):
            self.snap_image(elem, parent_transform)

    def effect(self):
        if self.options.snap_to == "bl":
            self.document_offset = (
                self.svg.to_dimensionless(self.svg.get_viewbox()[3]) % 1
            )
        else:
            self.document_offset = 0
        for id, elem in self.svg.selection.items():
            try:
                self.pixel_snap(elem)
            except TransformError as err:
                raise inkex.AbortExtension(str(err))


if __name__ == "__main__":
    PixelSnap().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2013 Public Domain
#               2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import sys
import hpgl_decoder
import inkex
from inkex.localization import inkex_gettext as _


class HpglInput(inkex.InputExtension):
    def add_arguments(self, pars):
        pars.add_argument(
            "--resolutionX", type=float, default=1016.0, help="Resolution X (dpi)"
        )
        pars.add_argument(
            "--resolutionY", type=float, default=1016.0, help="Resolution Y (dpi)"
        )
        pars.add_argument(
            "--showMovements",
            type=inkex.Boolean,
            default=False,
            help="Show Movements between paths",
        )

    def load(self, stream):
        return b";".join(line.strip() for line in stream).decode()

    def effect(self):
        # interpret HPGL data
        myHpglDecoder = hpgl_decoder.hpglDecoder(self.document, self.options)
        self.document = None

        try:
            doc, warnings = myHpglDecoder.get_svg()
        except Exception as inst:
            if inst.args[0] == "NO_HPGL_DATA":
                # issue error if no hpgl data found
                inkex.errormsg(_("No HPGL data found."))
                exit(1)
            else:
                type, value, traceback = sys.exc_info()
                raise ValueError("", type, value).with_traceback(traceback)

        # issue warning if unknown commands where found
        if "UNKNOWN_COMMANDS" in warnings:
            inkex.errormsg(
                _(
                    "The HPGL data contained unknown (unsupported) commands, "
                    "there is a possibility that the drawing is missing some content."
                )
            )

        # deliver document to inkscape
        self.document = doc


if __name__ == "__main__":
    HpglInput().run()



#!/usr/bin/env python
"""Reduce hue"""

import inkex


class LessHue(inkex.ColorExtension):
    """Remove Hue from the color"""

    def modify_color(self, name, color):
        color.hue -= int(0.05 * 255)
        return color


if __name__ == "__main__":
    LessHue().run()



# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import inkex
from inkex.transforms import DirectedLineSegment
from inkex.localization import inkex_gettext as _
from operator import truediv


class Envelope(inkex.EffectExtension):
    """Distort a path/group of paths to a second path"""

    def effect(self):
        if len(self.svg.selection) != 2:
            raise inkex.AbortExtension(_("You must select two objects only."))

        obj, envelope = self.svg.selection

        if isinstance(obj, (inkex.PathElement, inkex.Group)):
            if isinstance(envelope, inkex.PathElement):
                # Get bounding box plus any extra composed transform of parents.
                bbox = obj.bounding_box(obj.getparent().composed_transform())

                # distill trafo into four node points
                path = envelope.path.transform(
                    envelope.composed_transform()
                ).to_superpath()
                tbox = self.envelope_box_from_path(path)
            else:
                if isinstance(envelope, inkex.Group):
                    raise inkex.AbortExtension(
                        _(
                            "The second selected object is a group, not a"
                            " path.\nTry using Object->Ungroup."
                        )
                    )
                raise inkex.AbortExtension(
                    _(
                        "The second selected object is not a path.\nTry using"
                        " the procedure Path->Object to Path."
                    )
                )
        else:
            raise inkex.AbortExtension(
                _(
                    "The first selected object is neither a path nor a group.\nTry using"
                    " the procedure Path->Object to Path."
                )
            )

        self.process_object(obj, tbox, bbox)

    def envelope_box_from_path(self, envelope_path):
        if len(envelope_path) < 1 or len(envelope_path[0]) < 4:
            raise inkex.AbortExtension(
                _("Second selected path is too short. Must be four or more nodes.")
            )
        trafo = [[(csp[1][0], csp[1][1]) for csp in subs] for subs in envelope_path][0][
            :4
        ]
        # vectors pointing away from the trafo origin
        tbox = [
            DirectedLineSegment(trafo[0], trafo[1]),
            DirectedLineSegment(trafo[1], trafo[2]),
            DirectedLineSegment(trafo[3], trafo[2]),
            DirectedLineSegment(trafo[0], trafo[3]),
        ]
        vects = [segment.vector for segment in tbox]
        if (
            0.0
            == vects[0].cross(vects[1])
            == vects[1].cross(vects[2])
            == vects[2].cross(vects[3])
        ):
            raise inkex.AbortExtension(
                _("The points for the selected envelope must not all be in a line.")
            )
        return tbox

    def process_object(self, obj, tbox, bbox):
        if isinstance(obj, inkex.PathElement):
            self.process_path(obj, tbox, bbox)
        elif isinstance(obj, inkex.Group):
            self.process_group(obj, tbox, bbox)

    def process_group(self, group, tbox, bbox):
        """Go through all groups to process all paths inside them"""
        for node in group:
            self.process_object(node, tbox, bbox)

    def process_path(self, element, tbox, bbox):
        # Get out path's absolute and root coordinates, so obj and envelope
        # are always in the same coordinate system.
        points = (
            element.path.to_absolute()
            .transform(element.composed_transform())
            .to_superpath()
        )

        for subs in points:
            for csp in subs:
                csp[0] = self.transform_point(tbox, bbox, *csp[0])
                csp[1] = self.transform_point(tbox, bbox, *csp[1])
                csp[2] = self.transform_point(tbox, bbox, *csp[2])

        # Put the modified path back, undo the root transformation
        element.path = points.to_path().transform(-element.composed_transform())

    @staticmethod
    def transform_point(tbox, bbox, x, y):
        """Transform algorithm thanks to Jose Hevia (freon)"""
        vector = (x, y) - bbox.minimum
        xratio, yratio = map(truediv, vector, (bbox.width, bbox.height))
        horz = DirectedLineSegment(
            tbox[0].point_at_ratio(xratio), tbox[2].point_at_ratio(xratio)
        )
        vert = DirectedLineSegment(
            tbox[3].point_at_ratio(yratio), tbox[1].point_at_ratio(yratio)
        )
        denom = horz.vector.cross(vert.vector)
        if denom == 0.0:
            # Degenerate cases of intersecting envelope edges
            if horz.length == 0.0:
                return horz.start
            if vert.length == 0.0:
                return vert.start
            # Here we should know that the lines share a start or end point.
            if horz.vector.dot(vert.vector) < 0:
                # Segments point opposite directions
                if (horz.start - vert.start).length <= 1e-8:
                    return horz.start
                return horz.end
            # Otherwise they are pointing the same direction
            if (horz.start - vert.end).length < 1e-8:
                return horz.start
            return horz.end
        # If we didn't hit a degenerate case we can treat the segments as infinite lines
        intersect_ratio = (vert.start - horz.start).cross(vert.vector) / denom
        return horz.point_at_ratio(intersect_ratio)


if __name__ == "__main__":
    Envelope().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2008 Aurelio A. Heckert <aurium(a)gmail.com>
#    Week number option added by Olav Vitters and Nicolas Dufour (2012)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
A calendar generator plugin for Inkscape, but also can be used as a standalone
command line application.
#
More on ISO week number calculation on:
http://en.wikipedia.org/wiki/ISO_week_date
(The first week of a year is the week that contains the first Thursdayof the year.)
"""

__version__ = "0.4"

import datetime
import calendar
import re
import sys

import inkex
from inkex import TextElement
from inkex.localization import inkex_gettext as _

if sys.version_info[0] > 2:

    def unicode(stringlike, encoding):
        """Compatibility for python 2 strings"""
        if isinstance(stringlike, bytes):
            return stringlike.decode(encoding)
        return stringlike


class Calendar(inkex.EffectExtension):
    """Generate Calendar in SVG"""

    def add_arguments(self, pars):
        pars.add_argument("--tab", type=str, dest="tab")
        pars.add_argument(
            "--month",
            type=int,
            default=0,
            help="Month to be generated. If 0, then the entry year will be generated.",
        )
        pars.add_argument(
            "--year",
            type=int,
            default=0,
            help="Year to be generated. If 0, then the current year will be generated.",
        )
        pars.add_argument(
            "--fill-empty-day-boxes",
            type=inkex.Boolean,
            dest="fill_edb",
            default=True,
            help="Fill empty day boxes with next month days.",
        )
        pars.add_argument(
            "--show-week-number",
            type=inkex.Boolean,
            dest="show_weeknr",
            default=False,
            help="Include a week number column.",
        )
        pars.add_argument(
            "--start-day", default="sun", help='Week start day. ("sun" or "mon")'
        )
        pars.add_argument(
            "--weekend",
            default="sat+sun",
            help='Define the weekend days. ("sat+sun" or "sat" or "sun")',
        )
        pars.add_argument(
            "--auto-organize",
            type=inkex.Boolean,
            dest="auto_organize",
            default=True,
            help="Automatically set the size and positions.",
        )
        pars.add_argument(
            "--months-per-line",
            type=int,
            dest="months_per_line",
            default=3,
            help="Number of months side by side.",
        )
        pars.add_argument(
            "--month-width",
            type=str,
            dest="month_width",
            default="6cm",
            help="The width of the month days box.",
        )
        pars.add_argument(
            "--month-margin",
            type=str,
            dest="month_margin",
            default="1cm",
            help="The space between the month boxes.",
        )
        pars.add_argument(
            "--color-year",
            type=inkex.Color,
            dest="color_year",
            default=inkex.Color(0x808080FF),
            help="Color for the year header.",
        )
        pars.add_argument(
            "--color-month",
            type=inkex.Color,
            dest="color_month",
            default=inkex.Color(0x686868FF),
            help="Color for the month name header.",
        )
        pars.add_argument(
            "--color-day-name",
            type=inkex.Color,
            dest="color_day_name",
            default=inkex.Color(0x909090FF),
            help="Color for the week day names header.",
        )
        pars.add_argument(
            "--color-day",
            type=inkex.Color,
            dest="color_day",
            default=inkex.Color(0x000000FF),
            help="Color for the common day box.",
        )
        pars.add_argument(
            "--color-weekend",
            type=inkex.Color,
            dest="color_weekend",
            default=inkex.Color(0x787878FF),
            help="Color for the weekend days.",
        )
        pars.add_argument(
            "--color-nmd",
            type=inkex.Color,
            dest="color_nmd",
            default=inkex.Color(0xB0B0B0FF),
            help="Color for the next month day, in empty day boxes.",
        )
        pars.add_argument(
            "--color-weeknr",
            type=inkex.Color,
            dest="color_weeknr",
            default=inkex.Color(0x787878FF),
            help="Color for the week numbers.",
        )
        pars.add_argument(
            "--font-year",
            type=str,
            dest="font_year",
            default="arial",
            help="Font for the year string.",
        )
        pars.add_argument(
            "--font-month",
            type=str,
            dest="font_month",
            default="arial",
            help="Font for the month strings.",
        )
        pars.add_argument(
            "--font-day-name",
            type=str,
            dest="font_day_name",
            default="arial",
            help="Font for the days of the week strings.",
        )
        pars.add_argument(
            "--font-day",
            type=str,
            dest="font_day",
            default="arial",
            help="Font for the day strings.",
        )
        pars.add_argument(
            "--month-names",
            type=str,
            dest="month_names",
            default="January February March "
            "April May June July "
            "August September October "
            "November December",
            help="The month names for localization.",
        )
        pars.add_argument(
            "--day-names",
            type=str,
            dest="day_names",
            default="Sun Mon Tue Wed Thu Fri Sat",
            help="The week day names for localization.",
        )
        pars.add_argument(
            "--weeknr-name",
            type=str,
            dest="weeknr_name",
            default="Wk",
            help="The week number column name for localization.",
        )
        pars.add_argument(
            "--encoding",
            type=str,
            dest="input_encode",
            default="arabic",
            help="The input encoding of the names.",
        )

    def validate_options(self):
        """Validity check for various text inputs"""
        # inkex.errormsg( self.options.input_encode )
        # Convert string names lists in real lists
        month_name = re.match(r"\s*(.*[^\s])\s*", self.options.month_names)
        self.options.month_names = re.split(r"\s+", month_name.group(1))
        month_name = re.match(r"\s*(.*[^\s])\s*", self.options.day_names)
        self.options.day_names = re.split(r"\s+", month_name.group(1))

        # Validate names lists
        if len(self.options.month_names) != 12:
            inkex.errormsg(
                'The month name list "'
                + str(self.options.month_names)
                + '" is invalid. Using default.'
            )
            self.options.month_names = [
                "January",
                "February",
                "March",
                "April",
                "May",
                "June",
                "July",
                "August",
                "September",
                "October",
                "November",
                "December",
            ]
        if len(self.options.day_names) != 7:
            inkex.errormsg(
                'The day name list "'
                + str(self.options.day_names)
                + '" is invalid. Using default.'
            )
            self.options.day_names = ["Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"]
        # Convert year 0 to current year
        if self.options.year == 0:
            self.options.year = datetime.datetime.today().year
        # Year 1 starts it's week at monday, obligatorily
        if self.options.year == 1:
            self.options.start_day = "mon"
        # Set the calendar start day
        if self.options.start_day == "sun":
            calendar.setfirstweekday(6)
        else:
            calendar.setfirstweekday(0)
        # Convert string numbers with unit to user space float numbers
        self.options.month_width = self.svg.unittouu(self.options.month_width)
        self.options.month_margin = self.svg.unittouu(self.options.month_margin)

    # initial values
    month_x_pos = 0
    month_y_pos = 0
    weeknr = 0

    def calculate_size_and_positions(self):
        """month_margin month_width months_per_line auto_organize"""
        self.doc_w = self.svg.unittouu(self.document.getroot().get("width"))
        self.doc_h = self.svg.unittouu(self.document.getroot().get("height"))
        if self.options.show_weeknr:
            self.cols_before = 1
        else:
            self.cols_before = 0
        if self.options.auto_organize:
            if self.doc_h > self.doc_w:
                self.months_per_line = 3
            else:
                self.months_per_line = 4
        else:
            self.months_per_line = self.options.months_per_line
        # self.month_w = self.doc_w / self.months_per_line
        if self.options.auto_organize:
            self.month_w = (self.doc_w * 0.8) / self.months_per_line
            self.month_margin = self.month_w / 10
        else:
            self.month_w = self.options.month_width
            self.month_margin = self.options.month_margin
        self.day_w = self.month_w / (7 + self.cols_before)
        self.day_h = self.month_w / 9
        self.month_h = self.day_w * 7
        if self.options.month == 0:
            self.year_margin = (
                self.doc_w
                + self.day_w
                - (self.month_w * self.months_per_line)
                - (self.month_margin * (self.months_per_line - 1))
            ) / 2  # - self.month_margin
        else:
            self.year_margin = (self.doc_w - self.month_w) / 2
        self.style_day = {
            "font-size": str(self.day_w / 2),
            "font-family": self.options.font_day,
            "text-anchor": "middle",
            "text-align": "center",
            "fill": self.options.color_day,
        }
        self.style_weekend = self.style_day.copy()
        self.style_weekend["fill"] = self.options.color_weekend
        self.style_nmd = self.style_day.copy()
        self.style_nmd["fill"] = self.options.color_nmd
        self.style_month = self.style_day.copy()
        self.style_month["fill"] = self.options.color_month
        self.style_month["font-family"] = self.options.font_month
        self.style_month["font-size"] = str(self.day_w / 1.5)
        self.style_month["font-weight"] = "bold"
        self.style_day_name = self.style_day.copy()
        self.style_day_name["fill"] = self.options.color_day_name
        self.style_day_name["font-family"] = self.options.font_day_name
        self.style_day_name["font-size"] = str(self.day_w / 3)
        self.style_year = self.style_day.copy()
        self.style_year["fill"] = self.options.color_year
        self.style_year["font-family"] = self.options.font_year
        self.style_year["font-size"] = str(self.day_w * 2)
        self.style_year["font-weight"] = "bold"
        self.style_weeknr = self.style_day.copy()
        self.style_weeknr["fill"] = self.options.color_weeknr
        self.style_weeknr["font-size"] = str(self.day_w / 3)

    def is_weekend(self, pos):
        """Detect weekend days; weekend values: "sat+sun" or "sat" or "sun" """
        if self.options.start_day == "sun":
            if self.options.weekend == "sat+sun" and pos == 0:
                return True
            if self.options.weekend == "sat+sun" and pos == 6:
                return True
            if self.options.weekend == "sat" and pos == 6:
                return True
            if self.options.weekend == "sun" and pos == 0:
                return True
        else:
            if self.options.weekend == "sat+sun" and pos == 5:
                return True
            if self.options.weekend == "sat+sun" and pos == 6:
                return True
            if self.options.weekend == "sat" and pos == 5:
                return True
            if self.options.weekend == "sun" and pos == 6:
                return True
        return False

    def in_line_month(self, cal):
        """Lay out days of week in each week"""
        cal2 = []
        for week in cal:
            for day in week:
                if day != 0:
                    cal2.append(day)
        return cal2

    def write_month_header(self, g, m):
        """Write month header"""
        txt_atts = {
            "style": str(inkex.Style(self.style_month)),
            "x": str((self.month_w - self.day_w) / 2),
            "y": str(self.day_h / 5),
        }
        try:
            g.add(TextElement(**txt_atts)).text = unicode(
                self.options.month_names[m - 1], self.options.input_encode
            )
        except:
            raise ValueError(_("You must select a correct system encoding."))

        week_group = g.add(inkex.Group())
        week_x = 0
        if self.options.start_day == "sun":
            day_names = self.options.day_names[:]
        else:
            day_names = self.options.day_names[1:]
            day_names.append(self.options.day_names[0])

        if self.options.show_weeknr:
            day_names.insert(0, self.options.weeknr_name)

        for wday in day_names:
            txt_atts = {
                "style": str(inkex.Style(self.style_day_name)),
                "x": str(self.day_w * week_x),
                "y": str(self.day_h),
            }
            try:
                week_group.add(TextElement(**txt_atts)).text = unicode(
                    wday, self.options.input_encode
                )
            except:
                raise ValueError(_("You must select a correct system encoding."))

            week_x += 1

    def create_month(self, m):
        """Lay out one month in place on the calendar"""
        txt_atts = {
            "transform": "translate("
            + str(
                self.year_margin + (self.month_w + self.month_margin) * self.month_x_pos
            )
            + ","
            + str((self.day_h * 4) + (self.month_h * self.month_y_pos))
            + ")",
            "id": "month_" + str(m) + "_" + str(self.options.year),
        }
        g = self.year_g.add(inkex.Group(**txt_atts))
        self.write_month_header(g, m)
        gdays = g.add(inkex.Group())
        cal = calendar.monthcalendar(self.options.year, m)
        if m == 1:
            if self.options.year > 1:
                before_month = self.in_line_month(
                    calendar.monthcalendar(self.options.year - 1, 12)
                )
        else:
            before_month = self.in_line_month(
                calendar.monthcalendar(self.options.year, m - 1)
            )
        if m == 12:
            next_month = self.in_line_month(
                calendar.monthcalendar(self.options.year + 1, 1)
            )
        else:
            next_month = self.in_line_month(
                calendar.monthcalendar(self.options.year, m + 1)
            )
        if len(cal) < 6:
            # add a line after the last week
            cal.append([0, 0, 0, 0, 0, 0, 0])
        if len(cal) < 6:
            # add a line before the first week (Feb 2009)
            cal.reverse()
            cal.append([0, 0, 0, 0, 0, 0, 0])
            cal.reverse()
        # How mutch before month days will be showed:
        bmd = cal[0].count(0) + cal[1].count(0)
        before = True
        week_y = 0
        for week in cal:
            if (
                self.weeknr != 0
                and (
                    (self.options.start_day == "mon" and week[0] != 0)
                    or (self.options.start_day == "sun" and week[1] != 0)
                )
            ) or (
                self.weeknr == 0
                and (
                    (self.options.start_day == "mon" and week[3] > 0)
                    or (self.options.start_day == "sun" and week[4] > 0)
                )
            ):
                self.weeknr += 1
            week_x = 0
            if self.options.show_weeknr:
                # Remove leap week (starting previous year) and empty weeks
                if self.weeknr != 0 and not (week[0] == 0 and week[6] == 0):
                    style = self.style_weeknr
                    txt_atts = {
                        "style": str(inkex.Style(style)),
                        "x": str(self.day_w * week_x),
                        "y": str(self.day_h * (week_y + 2)),
                    }
                    gdays.add(TextElement(**txt_atts)).text = str(self.weeknr)
                    week_x += 1
                else:
                    week_x += 1
            for day in week:
                style = self.style_day
                if self.is_weekend(week_x - self.cols_before):
                    style = self.style_weekend
                if day == 0:
                    style = self.style_nmd
                txt_atts = {
                    "style": str(inkex.Style(style)),
                    "x": str(self.day_w * week_x),
                    "y": str(self.day_h * (week_y + 2)),
                }
                text = None
                if day == 0 and not self.options.fill_edb:
                    pass  # draw nothing
                elif day == 0:
                    if before:
                        text = str(before_month[-bmd])
                        bmd -= 1
                    else:
                        text = str(next_month[bmd])
                        bmd += 1
                else:
                    text = str(day)
                    before = False
                if text:
                    gdays.add(TextElement(**txt_atts)).text = text
                week_x += 1
            week_y += 1
        self.month_x_pos += 1
        if self.month_x_pos >= self.months_per_line:
            self.month_x_pos = 0
            self.month_y_pos += 1

    def effect(self):
        self.validate_options()
        self.calculate_size_and_positions()
        parent = self.document.getroot()
        txt_atts = {"id": "year_" + str(self.options.year)}
        self.year_g = parent.add(inkex.Group(**txt_atts))
        txt_atts = {
            "style": str(inkex.Style(self.style_year)),
            "x": str(self.doc_w / 2),
            "y": str(self.day_w * 1.5),
        }
        self.year_g.add(TextElement(**txt_atts)).text = str(self.options.year)
        try:
            if self.options.month == 0:
                for m in range(1, 13):
                    self.create_month(m)
            else:
                self.create_month(self.options.month)
        except ValueError as err:
            return inkex.errormsg(str(err))
        return None


if __name__ == "__main__":
    Calendar().run()
